from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import yaml
from torch.nn.modules.batchnorm import _BatchNorm

__all__ = [
    "make_divisible",
    "load_state_dict_from_file",
    "list_mean",
    "list_sum",
    "parse_unknown_args",
    "partial_update_config",
    "remove_bn",
    "get_same_padding",
    "torch_random_choices",
]


def make_divisible(
    v: Union[int, float], divisor: Optional[int], min_val=None
) -> Union[int, float]:
    """This function is taken from the original tf repo.

    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_val:
    :return:
    """
    if divisor is None:
        return v

    if min_val is None:
        min_val = divisor
    new_v = max(min_val, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def load_state_dict_from_file(file: str) -> Dict[str, torch.Tensor]:
    checkpoint = torch.load(file, map_location="cpu")
    if "state_dict" in checkpoint:
        checkpoint = checkpoint["state_dict"]
    return checkpoint


def list_sum(x: List) -> Any:
    return x[0] if len(x) == 1 else x[0] + list_sum(x[1:])


def list_mean(x: List) -> Any:
    return list_sum(x) / len(x)


def parse_unknown_args(unknown: List) -> Dict:
    """Parse unknown args."""
    index = 0
    parsed_dict = {}
    while index < len(unknown):
        key, val = unknown[index], unknown[index + 1]
        index += 2
        if key.startswith("--"):
            key = key[2:]
            try:
                # try parsing with yaml
                if "{" in val and "}" in val and ":" in val:
                    val = val.replace(":", ": ")  # add space manually for dict
                out_val = yaml.safe_load(val)
            except ValueError:
                # return raw string if parsing fails
                out_val = val
            parsed_dict[key] = out_val
    return parsed_dict


def partial_update_config(config: Dict, partial_config: Dict):
    for key in partial_config:
        if (
            key in config
            and isinstance(partial_config[key], Dict)
            and isinstance(config[key], Dict)
        ):
            partial_update_config(config[key], partial_config[key])
        else:
            config[key] = partial_config[key]


def remove_bn(model: nn.Module) -> None:
    for m in model.modules():
        if isinstance(m, _BatchNorm):
            m.weight = m.bias = None
            m.forward = lambda x: x


def get_same_padding(kernel_size: Union[int, Tuple[int, int]]) -> Union[int, tuple]:
    if isinstance(kernel_size, tuple):
        assert len(kernel_size) == 2, f"invalid kernel size: {kernel_size}"
        p1 = get_same_padding(kernel_size[0])
        p2 = get_same_padding(kernel_size[1])
        return p1, p2
    else:
        assert isinstance(
            kernel_size, int
        ), "kernel size should be either `int` or `tuple`"
        assert kernel_size % 2 > 0, "kernel size should be odd number"
        return kernel_size // 2


def torch_random_choices(
    src_list: List[Any],
    generator: Optional[torch.Generator],
    k=1,
) -> Union[Any, List[Any]]:
    rand_idx = torch.randint(low=0, high=len(src_list), generator=generator, size=(k,))
    out_list = [src_list[i] for i in rand_idx]
    return out_list[0] if k == 1 else out_list


from typing import List, Optional, Union

import torch
import torch.distributed
from torchpack import distributed

from rrelu.utils.misc import list_mean, list_sum

__all__ = ["ddp_reduce_tensor", "DistributedMetric"]
def ddp_reduce_tensor(
    tensor: torch.Tensor, reduce="mean"
) -> Union[torch.Tensor, List[torch.Tensor]]:
    tensor_list = [torch.empty_like(tensor) for _ in range(distributed.size())]
    torch.distributed.all_gather(tensor_list, tensor.contiguous(), async_op=False)
    if reduce == "mean":
        return list_mean(tensor_list)
    elif reduce == "sum":
        return list_sum(tensor_list)
    elif reduce == "cat":
        return torch.cat(tensor_list, dim=0)
    elif reduce == "root":
        return tensor_list[0]
    else:
        return tensor_list


class DistributedMetric(object):
    """Average metrics for distributed training."""

    def __init__(self, name: Optional[str] = None, backend="ddp"):
        self.name = name
        self.sum = 0
        self.count = 0
        self.backend = backend

    def update(self, val: Union[torch.Tensor, int, float], delta_n=1):
        val *= delta_n
        if type(val) in [int, float]:
            val = torch.Tensor(1).fill_(val).cuda()
        if self.backend == "ddp":
            self.count += ddp_reduce_tensor(
                torch.Tensor(1).fill_(delta_n).cuda(), reduce="sum"
            )
            self.sum += ddp_reduce_tensor(val.detach(), reduce="sum")
        else:
            raise NotImplementedError

    @property
    def avg(self):
        if self.count == 0:
            return torch.Tensor(1).fill_(-1)
        else:
            return self.sum / self.count


"""pytorchfi.error_models provides different error models out-of-the-box for use."""

import logging
import random

import torch

from rrelu.pytorchfi import core
from rrelu.pytorchfi.util import random_value

# Helper Functions


def random_batch_element(pfi: core.FaultInjection):
    return random.randint(0, pfi.batch_size - 1)


def random_neuron_location(pfi: core.FaultInjection, layer: int = -1):
    if layer == -1:
        layer = random.randint(0, pfi.get_total_layers() - 1)

    dim = pfi.get_layer_dim(layer)
    shape = pfi.get_layer_shape(layer)

    dim1_shape = shape[1]
    dim1_rand = random.randint(0, dim1_shape - 1)
    if dim > 2:
        dim2_shape = shape[2]
        dim2_rand = random.randint(0, dim2_shape - 1)
    else:
        dim2_rand = None
    if dim > 3:
        dim3_shape = shape[3]
        dim3_rand = random.randint(0, dim3_shape - 1)
    else:
        dim3_rand = None

    return (layer, dim1_rand, dim2_rand, dim3_rand)


# Neuron Perturbation Models

# single random neuron error in single batch element
def random_neuron_inj(pfi: core.FaultInjection, min_val: int = -1, max_val: int = 1):
    b = random_batch_element(pfi)
    (layer, C, H, W) = random_neuron_location(pfi)
    err_val = random_value(min_val=min_val, max_val=max_val)

    return pfi.declare_neuron_fault_injection(
        batch=[b], layer_num=[layer], dim1=[C], dim2=[H], dim3=[W], value=[err_val]
    )


# single random neuron error in each batch element.
def random_neuron_inj_batched(
    pfi: core.FaultInjection,
    min_val: int = -1,
    max_val: int = 1,
    rand_loc: bool = True,
    rand_val: bool = True,
):
    batch, layer_num, c_rand, h_rand, w_rand, value = ([] for _ in range(6))

    if not rand_loc:
        (layer, C, H, W) = random_neuron_location(pfi)
    if not rand_val:
        err_val = random_value(min_val=min_val, max_val=max_val)

    for i in range(pfi.batch_size):
        if rand_loc:
            (layer, C, H, W) = random_neuron_location(pfi)
        if rand_val:
            err_val = random_value(min_val=min_val, max_val=max_val)

        batch.append(i)
        layer_num.append(layer)
        c_rand.append(C)
        h_rand.append(H)
        w_rand.append(W)
        value.append(err_val)

    return pfi.declare_neuron_fault_injection(
        batch=batch,
        layer_num=layer_num,
        dim1=c_rand,
        dim2=h_rand,
        dim3=w_rand,
        value=value,
    )


# one random neuron error per layer in single batch element
def random_inj_per_layer(pfi: core.FaultInjection, min_val: int = -1, max_val: int = 1):
    batch, layer_num, c_rand, h_rand, w_rand, value = ([] for i in range(6))

    b = random_batch_element(pfi)
    for i in range(pfi.get_total_layers()):
        (layer, C, H, W) = random_neuron_location(pfi, layer=i)
        batch.append(b)
        layer_num.append(layer)
        c_rand.append(C)
        h_rand.append(H)
        w_rand.append(W)
        value.append(random_value(min_val=min_val, max_val=max_val))

    return pfi.declare_neuron_fault_injection(
        batch=batch,
        layer_num=layer_num,
        dim1=c_rand,
        dim2=h_rand,
        dim3=w_rand,
        value=value,
    )


# one random neuron error per layer in each batch element
def random_inj_per_layer_batched(
    pfi: core.FaultInjection,
    min_val: int = -1,
    max_val: int = 1,
    rand_loc: bool = True,
    rand_val: bool = True,
):
    batch, layer_num, c_rand, h_rand, w_rand, value = ([] for i in range(6))

    for i in range(pfi.get_total_layers()):
        if not rand_loc:
            (layer, C, H, W) = random_neuron_location(pfi, layer=i)
        if not rand_val:
            err_val = random_value(min_val=min_val, max_val=max_val)

        for b in range(pfi.batch_size):
            if rand_loc:
                (layer, C, H, W) = random_neuron_location(pfi, layer=i)
            if rand_val:
                err_val = random_value(min_val=min_val, max_val=max_val)

            batch.append(b)
            layer_num.append(layer)
            c_rand.append(C)
            h_rand.append(H)
            w_rand.append(W)
            value.append(err_val)

    return pfi.declare_neuron_fault_injection(
        batch=batch,
        layer_num=layer_num,
        dim1=c_rand,
        dim2=h_rand,
        dim3=w_rand,
        value=value,
    )


class single_bit_flip_func(core.FaultInjection):
    def __init__(self, model, batch_size, input_shape=None, **kwargs):
        if input_shape is None:
            input_shape = [3, 224, 224]
        super().__init__(model, batch_size, input_shape=input_shape, **kwargs)
        logging.basicConfig(format="%(asctime)-15s %(clientip)s %(user)-8s %(message)s")

        self.bits = kwargs.get("bits", 8)
        self.layer_ranges = []

    def set_conv_max(self, data):
        self.layer_ranges = data

    def reset_conv_max(self, data):
        self.layer_ranges = []

    def get_conv_max(self, layer):
        return self.layer_ranges[layer]

    @staticmethod
    def _twos_comp(val, bits):
        if (val & (1 << (bits - 1))) != 0:
            val = val - (1 << bits)
        return val

    def _twos_comp_shifted(self, val, nbits):
        return (1 << nbits) + val if val < 0 else self._twos_comp(val, nbits)

    def _flip_bit_signed(self, orig_value, max_value, bit_pos):
        # quantum value
        save_type = orig_value.dtype
        total_bits = self.bits
        logging.info(f"Original Value: {orig_value}")

        quantum = int((orig_value / max_value) * ((2.0 ** (total_bits - 1))))
        twos_comple = self._twos_comp_shifted(quantum, total_bits)  # signed
        logging.info(f"Quantum: {quantum}")
        logging.info(f"Twos Couple: {twos_comple}")

        # binary representation
        bits = bin(twos_comple)[2:]
        logging.info(f"Bits: {bits}")

        # sign extend 0's
        temp = "0" * (total_bits - len(bits))
        bits = temp + bits
        if len(bits) != total_bits:
            raise AssertionError
        logging.info(f"Sign extend bits {bits}")

        # flip a bit
        # use MSB -> LSB indexing
        if bit_pos >= total_bits:
            raise AssertionError

        bits_new = list(bits)
        bit_loc = total_bits - bit_pos - 1
        if bits_new[bit_loc] == "0":
            bits_new[bit_loc] = "1"
        else:
            bits_new[bit_loc] = "0"
        bits_str_new = "".join(bits_new)
        logging.info(f"New bits: {bits_str_new}")

        # GPU contention causes a weird bug...
        if not bits_str_new.isdigit():
            logging.info("Error: Not all the bits are digits (0/1)")

        # convert to quantum
        if not bits_str_new.isdigit():
            raise AssertionError
        new_quantum = int(bits_str_new, 2)
        out = self._twos_comp(new_quantum, total_bits)
        logging.info(f"Out: {out}")

        # get FP equivalent from quantum
        new_value = out * ((2.0 ** (-1 * (total_bits - 1))) * max_value)
        logging.info(f"New Value: {new_value}")

        return torch.tensor(new_value, dtype=save_type)

    def single_bit_flip_signed_across_batch(self, module, input_val, output):
        corrupt_conv_set = self.corrupt_layer
        range_max = self.get_conv_max(self.current_layer)
        logging.info(f"Current layer: {self.current_layer}")
        logging.info(f"Range_max: {range_max}")

        if type(corrupt_conv_set) is list:
            inj_list = list(
                filter(
                    lambda x: corrupt_conv_set[x] == self.current_layer,
                    range(len(corrupt_conv_set)),
                )
            )
            for i in inj_list:
                self.assert_injection_bounds(index=i)
                prev_value = output[self.corrupt_batch[i]][self.corrupt_dim[0][i]][
                    self.corrupt_dim[1][i]
                ][self.corrupt_dim[2][i]]

                rand_bit = random.randint(0, self.bits - 1)
                logging.info(f"Random Bit: {rand_bit}")
                new_value = self._flip_bit_signed(prev_value, range_max, rand_bit)

                output[self.corrupt_batch[i]][self.corrupt_dim[0][i]][
                    self.corrupt_dim[1][i]
                ][self.corrupt_dim[2][i]] = new_value

        else:
            if self.current_layer == corrupt_conv_set:
                prev_value = output[self.corrupt_batch][self.corrupt_dim[0]][
                    self.corrupt_dim[1]
                ][self.corrupt_dim[2]]

                rand_bit = random.randint(0, self.bits - 1)
                logging.info(f"Random Bit: {rand_bit}")
                new_value = self._flip_bit_signed(prev_value, range_max, rand_bit)

                output[self.corrupt_batch][self.corrupt_dim[0]][self.corrupt_dim[1]][
                    self.corrupt_dim[2]
                ] = new_value

        self.update_layer()
        if self.current_layer >= len(self.output_size):
            self.reset_current_layer()


def random_neuron_single_bit_inj_batched(
    pfi: core.FaultInjection, layer_ranges, batch_random=True
):
    """
    Args:
        pfi: The core.FaultInjection in which the neuron fault injection should be instantiated.
        layer_ranges:
        batch_random (default True): True if each batch should have a random location, if false, then each
                                     batch will use the same randomly generated location.
    """
    pfi.set_conv_max(layer_ranges)

    locations = (
        [random_neuron_location(pfi) for _ in range(pfi.batch_size)]
        if batch_random
        else [random_neuron_location(pfi)] * pfi.batch_size
    )
    # Convert list of tuples [(1, 3), (2, 4)] to list of list [[1, 2], [3, 4]]
    random_layers, random_c, random_h, random_w = map(list, zip(*locations))

    return pfi.declare_neuron_fault_injection(
        batch=range(pfi.batch_size),
        layer_num=random_layers,
        dim1=random_c,
        dim2=random_h,
        dim3=random_w,
        function=pfi.single_bit_flip_signed_across_batch,
    )


def random_neuron_single_bit_inj(pfi: core.FaultInjection, layer_ranges):
    # TODO Support multiple error models via list
    pfi.set_conv_max(layer_ranges)

    batch = random_batch_element(pfi)
    (layer, C, H, W) = random_neuron_location(pfi)

    return pfi.declare_neuron_fault_injection(
        batch=[batch],
        layer_num=[layer],
        dim1=[C],
        dim2=[H],
        dim3=[W],
        function=pfi.single_bit_flip_signed_across_batch,
    )



"""pytorchfi.core contains the core functionality for fault injections"""

import copy
import logging
import warnings
from typing import List

import torch
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound

class FaultInjection:
    def __init__(
        self,
        model,
        batch_size: int,
        input_shape: List[int] = None,
        layer_types=None,
        total_bits = 32, 
        n_frac = 16,
        n_int = 15, 
        **kwargs,
    ):
        if not input_shape:
            input_shape = [3, 224, 224]
        if not layer_types:
            layer_types = [nn.Conv2d]
        logging.basicConfig(format="%(asctime)-15s %(clientip)s %(user)-8s %(message)s")

        self.original_model = model.cuda()
        self.output_size = []
        self.layers_type = []
        self.layers_dim = []
        self.weights_size = []
        self.bias_size= []
        self.batch_size = batch_size
        self._input_shape = input_shape
        self._inj_layer_types = layer_types
        self.total_bits = total_bits
        self.n_frac = n_frac
        self.n_int = n_int
        self.corrupted_model = None
        self.current_layer = 0
        self.handles = []
        self.corrupt_batch = []
        self.corrupt_layer = []
        self.corrupt_dim = [[], [], []]  # C, H, W
        self.corrupt_value = []

        self.use_cuda = kwargs.get("use_cuda", next(model.parameters()).is_cuda)

        if not isinstance(input_shape, list):
            raise AssertionError("Error: Input shape must be provided as a list.")
        if not (isinstance(batch_size, int) and batch_size >= 1):
            raise AssertionError("Error: Batch size must be an integer greater than 1.")
        if len(layer_types) < 0:
            raise AssertionError("Error: At least one layer type must be selected.")

        handles, _shapes, self.weights_size,self.bias_size = self._traverse_model_set_hooks(
            self.original_model, self._inj_layer_types
        )

        dummy_shape = (1, *self._input_shape)  # profiling only needs one batch element
        model_dtype = next(model.parameters()).dtype
        device = "cuda" if self.use_cuda else None
        _dummy_tensor = torch.randn(dummy_shape, dtype=model_dtype, device=device)

        self.original_model(_dummy_tensor)

        for index, _handle in enumerate(handles):
            handles[index].remove()

        logging.info("Input shape:")
        logging.info(dummy_shape[1:])

        logging.info("Model layer sizes:")
        logging.info(
            "\n".join(
                [
                    "".join(["{:4}".format(item) for item in row])
                    for row in self.output_size
                ]
            )
        )

    def reset_fault_injection(self):
        self._reset_fault_injection_state()
        self.corrupted_model = None
        logging.info("Fault injector reset.")

    def _reset_fault_injection_state(self):
        (
            self.current_layer,
            self.corrupt_batch,
            self.corrupt_layer,
            self.corrupt_dim,
            self.corrupt_value,
        ) = (0, [], [], [[], [], []], [])

        for index, _handle in enumerate(self.handles):
            self.handles[index].remove()

    def _traverse_model_set_hooks(self, model, layer_types):
        handles = []
        output_shape = []
        weights_shape = []
        bias_shape =[]
        for layer in model.children():
            # leaf node
            if list(layer.children()) == []:
                if "all" in layer_types:
                    handles.append(layer.register_forward_hook(self._save_output_size))
                else:
                    for i in layer_types:
                        if isinstance(layer, i):

                            # neurons
                            handles.append(
                                layer.register_forward_hook(self._save_output_size)
                            )
                            output_shape.append(layer)

                            # weights
                            if isinstance(layer,Relu_bound):
                                if len(layer.bounds.shape) == 0 : 
                                    weights_shape.append([0])
                                else:
                                    weights_shape.append(layer.bounds.shape)    
                                if layer.tresh!=None:
                                    if len(layer.tresh.shape)==0:
                                        bias_shape.append([0])
                                    else:
                                        bias_shape.append(layer.tresh.shape)    
                                else:
                                    bias_shape.append([None])
                            else:
                                weights_shape.append(layer.weight.shape)
                                if layer.bias!=None:
                                    bias_shape.append(layer.bias.shape)
                                else:
                                    bias_shape.append([None])           

            # unpack node
            else:
                subhandles, subbase, subweight,subbias = self._traverse_model_set_hooks(
                    layer, layer_types
                )
                for i in subhandles:
                    handles.append(i)
                for i in subbase:
                    output_shape.append(i)
                for i in subweight:
                    weights_shape.append(i)
                for i in subbias:
                    bias_shape.append(i)   

        return (handles, output_shape, weights_shape,bias_shape)

    def _traverse_model_set_hooks_neurons(self, model, layer_types, customInj, injFunc):
        handles = []
        for layer in model.children():
            # leaf node
            if list(layer.children()) == []:
                if "all" in layer_types:
                    hook = injFunc if customInj else self._set_value
                    handles.append(layer.register_forward_hook(hook))
                else:
                    for i in layer_types:
                        if isinstance(layer, i):
                            hook = injFunc if customInj else self._set_value
                            handles.append(layer.register_forward_hook(hook))
            # unpack node
            else:
                subHandles = self._traverse_model_set_hooks_neurons(
                    layer, layer_types, customInj, injFunc
                )
                for i in subHandles:
                    handles.append(i)

        return handles

    def declare_weight_fault_injection(self, **kwargs):
        self._reset_fault_injection_state()
        custom_injection = False
        custom_function = False

        if kwargs:
            if "function" in kwargs:
                custom_injection, custom_function = True, kwargs.get("function")
                corrupt_layer = kwargs.get("layer_num", [])
                corrupt_k = kwargs.get("k", [])
                corrupt_c = kwargs.get("dim1", [])
                corrupt_kH = kwargs.get("dim2", [])
                corrupt_kW = kwargs.get("dim3", [])
                corrupt_bit = kwargs.get("dim4",[])
                total_bits = kwargs.get("total_bits")
                n_frac = kwargs.get("n_frac")
                n_int = kwargs.get("n_int")
            else:
                corrupt_layer = kwargs.get(
                    "layer_num",
                )
                corrupt_k = kwargs.get("k", [])
                corrupt_c = kwargs.get("dim1", [])
                corrupt_kH = kwargs.get("dim2", [])
                corrupt_kW = kwargs.get("dim3", [])
                corrupt_value = kwargs.get("value", [])
                corrupt_bit = kwargs.get("dim4",[])
                total_bits = kwargs.get("total_bits")
                n_frac = kwargs.get("n_frac")
                n_int = kwargs.get("n_int")
        else:
            raise ValueError("Please specify an injection or injection function")

        # TODO: bound check here

        self.corrupted_model = copy.deepcopy(self.original_model)

        current_weight_layer = 0
        for layer in self.corrupted_model.modules():
            if isinstance(layer, tuple(self._inj_layer_types)):
                inj_list_weight = list(
                    filter(
                        lambda x: corrupt_layer[0][x] == current_weight_layer,
                        range(len(corrupt_layer[0])),
                    )
                )
                inj_list_bias = list(
                    filter(
                        lambda x: corrupt_layer[1][x] == current_weight_layer,
                        range(len(corrupt_layer[1])),
                    )
                )
                for inj in inj_list_weight:

                    corrupt_idx = tuple(
                        [
                            corrupt_k[0][inj],
                            corrupt_c[0][inj],
                            corrupt_kH[0][inj],
                            corrupt_kW[0][inj],
                        ]
                    )
                    # print(corrupt_idx)
                    corrupt_bit_idx = corrupt_bit[0][inj]
                    if isinstance(layer,Relu_bound):
                            orig_value = layer.bounds[corrupt_idx].item()
                            with torch.no_grad():
                                if custom_injection:
                                    corrupt_value = custom_function[1](layer.bounds, corrupt_idx,corrupt_bit_idx ,total_bits ,n_frac , n_int)
                                    layer.bounds[corrupt_idx].copy_(corrupt_value)
                                else:
                                    layer.bounds[corrupt_idx].copy_(corrupt_value[inj]) 
                    else:
                        orig_value = layer.weight[corrupt_idx].item()   
                        with torch.no_grad():
                            if custom_injection:
                                corrupt_value = custom_function[0](layer.weight, corrupt_idx,corrupt_bit_idx,total_bits ,n_frac , n_int)
                                # print(corrupt_value)
                                layer.weight[corrupt_idx].copy_(corrupt_value)
                            else:
                                layer.weight[corrupt_idx].copy_(corrupt_value[inj])

                for inj in inj_list_bias:
                    corrupt_idx = tuple(
                        [
                            corrupt_k[1][inj],
                            corrupt_c[1][inj],
                            corrupt_kH[1][inj],
                            corrupt_kW[1][inj],
                        ]
                    )
                    corrupt_bit_idx = corrupt_bit[1][inj]
                    # print(corrupt_idx)
                    if  isinstance(layer,Relu_bound):
                        orig_value = layer.tresh[corrupt_idx].item()
                        with torch.no_grad():
                            if custom_injection:
                                corrupt_value = custom_function[1](layer.tresh, corrupt_idx ,corrupt_bit_idx ,total_bits,n_frac , n_int)
                                layer.tresh[corrupt_idx].copy_(corrupt_value)
                            else:
                                layer.tresh[corrupt_idx].copy_(corrupt_value[inj])
                    else:    
                        orig_value = layer.bias[corrupt_idx].item()   
                        with torch.no_grad():
                            if custom_injection:
                                corrupt_value = custom_function[1](layer.bias, corrupt_idx ,corrupt_bit_idx , total_bits,n_frac , n_int)
                                layer.bias[corrupt_idx].copy_(corrupt_value)
                            else:
                                layer.bias[corrupt_idx].copy_(corrupt_value[inj])

                current_weight_layer += 1
        return self.corrupted_model

    def declare_neuron_fault_injection(self, **kwargs):
        self._reset_fault_injection_state()
        custom_injection = False
        injection_function = False

        if kwargs:
            if "function" in kwargs:
                logging.info("Declaring Custom Function")
                custom_injection, injection_function = True, kwargs.get("function")
            else:
                logging.info("Declaring Specified Fault Injector")
                self.corrupt_value = kwargs.get("value", [])

            self.corrupt_layer = kwargs.get("layer_num", [])
            self.corrupt_batch = kwargs.get("batch", [])
            self.corrupt_dim[0] = kwargs.get("dim1", [])
            self.corrupt_dim[1] = kwargs.get("dim2", [])
            self.corrupt_dim[2] = kwargs.get("dim3", [])

            logging.info(f"Convolution: {self.corrupt_layer}")
            logging.info("Batch, x, y, z:")
            logging.info(
                f"{self.corrupt_batch}, {self.corrupt_dim[0]}, {self.corrupt_dim[1]}, {self.corrupt_dim[2]}"
            )
        else:
            raise ValueError("Please specify an injection or injection function")

        self.check_bounds(
            self.corrupt_batch,
            self.corrupt_layer,
            self.corrupt_dim,
        )

        self.corrupted_model = copy.deepcopy(self.original_model)
        handles_neurons = self._traverse_model_set_hooks_neurons(
            self.corrupted_model,
            self._inj_layer_types,
            custom_injection,
            injection_function,
        )

        for i in handles_neurons:
            self.handles.append(i)

        return self.corrupted_model

    def check_bounds(self, batch, layer, dim):
        if (
            len(batch) != len(layer)
            or len(batch) != len(dim[0])
            or len(batch) != len(dim[1])
            or len(batch) != len(dim[2])
        ):
            raise AssertionError("Injection location missing values.")

        logging.info("Checking bounds before runtime")
        for i in range(len(batch)):
            self.assert_injection_bounds(i)

    def assert_injection_bounds(self, index: int):
        if index < 0:
            raise AssertionError(f"Invalid injection index: {index}")
        if self.corrupt_batch[index] >= self.batch_size:
            raise AssertionError(
                f"{self.corrupt_batch[index]} < {self.batch_size()}: Invalid batch element!"
            )
        if self.corrupt_layer[index] >= len(self.output_size):
            raise AssertionError(
                f"{self.corrupt_layer[index]} < {len(self.output_size)}: Invalid layer!"
            )

        corrupt_layer_num = self.corrupt_layer[index]
        layer_type = self.layers_type[corrupt_layer_num]
        layer_dim = self.layers_dim[corrupt_layer_num]
        layer_shape = self.output_size[corrupt_layer_num]

        for d in range(1, 4):
            if layer_dim > d and self.corrupt_dim[d - 1][index] >= layer_shape[d]:
                raise AssertionError(
                    f"{self.corrupt_dim[d - 1][index]} < {layer_shape[d]}: Out of bounds error in Dimension {d}!"
                )

        if layer_dim <= 2 and (
            self.corrupt_dim[1][index] is not None
            or self.corrupt_dim[2][index] is not None
        ):
            warnings.warn(
                f"Values in Dim2 and Dim3 ignored, since layer is {layer_type}"
            )

        if layer_dim <= 3 and self.corrupt_dim[2][index] is not None:
            warnings.warn(f"Values Dim3 ignored, since layer is {layer_type}")

        logging.info(f"Finished checking bounds on inj '{index}'")

    def _set_value(self, module, input_val, output):
        logging.info(
            f"Processing hook of Layer {self.current_layer}: {self.layers_type[self.current_layer]}"
        )
        inj_list = list(
            filter(
                lambda x: self.corrupt_layer[x] == self.current_layer,
                range(len(self.corrupt_layer)),
            )
        )

        layer_dim = self.layers_dim[self.current_layer]

        logging.info(f"Layer {self.current_layer} injection list size: {len(inj_list)}")
        if layer_dim == 2:
            for i in inj_list:
                self.assert_injection_bounds(index=i)
                logging.info(
                    f"Original value at [{self.corrupt_batch[i]}][{self.corrupt_dim[0][i]}]: {output[self.corrupt_batch[i]][self.corrupt_dim[0][i]]}"
                )
                logging.info(f"Changing value to {self.corrupt_value[i]}")
                output[self.corrupt_batch[i]][
                    self.corrupt_dim[0][i]
                ] = self.corrupt_value[i]
        elif layer_dim == 3:
            for i in inj_list:
                self.assert_injection_bounds(index=i)
                logging.info(
                    f"Original value at [{self.corrupt_batch[i]}][{self.corrupt_dim[0][i]}][{self.corrupt_dim[1][i]}]: {output[self.corrupt_batch[i]][self.corrupt_dim[0][i]][self.corrupt_dim[1][i]]}"
                )
                logging.info(f"Changing value to {self.corrupt_value[i]}")
                output[self.corrupt_batch[i]][
                    self.corrupt_dim[0][i], self.corrupt_dim[1][i]
                ] = self.corrupt_value[i]
        elif layer_dim == 4:
            for i in inj_list:
                self.assert_injection_bounds(index=i)
                logging.info(
                    f"Original value at [{self.corrupt_batch[i]}][{self.corrupt_dim[0][i]}][{self.corrupt_dim[1][i]}][{self.corrupt_dim[2][i]}]: {output[self.corrupt_batch[i]][self.corrupt_dim[0][i]][self.corrupt_dim[1][i]][self.corrupt_dim[2][i]]}"
                )
                logging.info(f"Changing value to {self.corrupt_value[i]}")
                output[self.corrupt_batch[i]][self.corrupt_dim[0][i]][
                    self.corrupt_dim[1][i]
                ][self.corrupt_dim[2][i]] = self.corrupt_value[i]

        self.update_layer()

    def _save_output_size(self, module, input_val, output):
        shape = list(output.size())
        dim = len(shape)

        self.layers_type.append(type(module))
        self.layers_dim.append(dim)
        self.output_size.append(shape)

    def update_layer(self, value=1):
        self.current_layer += value

    def reset_current_layer(self):
        self.current_layer = 0

    def get_weights_size(self, layer_num):
        return self.weights_size[layer_num]
    def get_bias_size(self,layer_num):
        return self.bias_size[layer_num]
    def get_weights_dim(self, layer_num):
        return len(self.weights_size[layer_num])

    def get_layer_type(self, layer_num):
        return self.layers_type[layer_num]

    def get_layer_dim(self, layer_num):
        return self.layers_dim[layer_num]

    def get_layer_shape(self, layer_num):
        return self.output_size[layer_num]
    def get_total_bits(self):
        return self.total_bits , self.n_frac,self.n_int
    def get_total_layers(self):
        return len(self.output_size)

    def get_tensor_dim(self, layer, dim):
        if dim > len(self.layers_dim):
            raise AssertionError(f"Dimension {dim} is out of bounds for layer {layer}")
        return self.output_size[layer][dim]

    def print_pytorchfi_layer_summary(self):
        summary_str = (
            "============================ PYTORCHFI INIT SUMMARY =============================="
            + "\n\n"
        )

        summary_str += "Layer types allowing injections:\n"
        summary_str += (
            "----------------------------------------------------------------------------------"
            + "\n"
        )
        for l_type in self._inj_layer_types:
            summary_str += "{:>5}".format("- ")
            substring = str(l_type).split(".")[-1].split("'")[0]
            summary_str += substring + "\n"
        summary_str += "\n"

        summary_str += "Model Info:\n"
        summary_str += (
            "----------------------------------------------------------------------------------"
            + "\n"
        )

        summary_str += "   - Shape of input into the model: ("
        for dim in self._input_shape:
            summary_str += str(dim) + " "
        summary_str += ")\n"

        summary_str += "   - Batch Size: " + str(self.batch_size) + "\n"
        summary_str += "   - CUDA Enabled: " + str(self.use_cuda) + "\n\n"

        summary_str += "Layer Info:\n"
        summary_str += (
            "----------------------------------------------------------------------------------"
            + "\n"
        )
        line_new = "{:>5}  {:>35}  {:>10} {:>20} {:>20} {:>20}".format(
            "Layer #", "Layer type", "Dimensions", "Weight Shape", "Output Shape","Bias Shape"
        )
        summary_str += line_new + "\n"
        summary_str += (
            "----------------------------------------------------------------------------------"
            + "\n"
        )
        for layer, _dim in enumerate(self.output_size):
            line_new = "{:>5}  {:>35}  {:>10} {:>20} {:>20} {:>20}".format(
                layer,
                str(self.layers_type[layer]).split(".")[-1].split("'")[0],
                str(self.layers_dim[layer]),
                str(list(self.weights_size[layer])),
                str(self.output_size[layer]),
                str(list(self.bias_size[layer]))
            )
            summary_str += line_new + "\n"

        summary_str += (
            "=================================================================================="
            + "\n"
        )

        logging.info(summary_str)
        return summary_str



import torch.optim as optim
import torchvision.transforms as transforms
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound
class AlexNet_model(nn.Module):
    def __init__(self, n_classes=10,dropout_rate=0.0):
        super(AlexNet_model, self).__init__()
        self.conv1 =  nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1)
        self.relu1 = nn.ReLU()
        self.maxpool1 =  nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(64, 192, kernel_size=3, padding=2)
        self.relu2 = nn.ReLU()
        self.maxpool2 = nn.MaxPool2d(kernel_size=2)
        self.conv3 = nn.Conv2d(192, 384, kernel_size=3, padding=1)
        self.relu3 = nn.ReLU()
        self.conv4 = nn.Conv2d(384, 256, kernel_size=3, padding=1)
        self.relu4 = nn.ReLU()
        self.conv5 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.relu5 = nn.ReLU()
        self.maxpool3 =  nn.MaxPool2d(kernel_size=2)
        # self.dropout = nn.Dropout(dropout_rate)
        self.linear1 = nn.Linear(256 * 2 * 2, 4096)
        self.relu6 = nn.ReLU()
        self.linear2 = nn.Linear(4096, 4096)
        self.relu7_last= nn.ReLU()
        self.linear3 = nn.Linear(4096, n_classes)
        # self.bn8 = nn.BatchNorm2d(10),
        # self.relu8_last= nn.ReLU()
    def forward(self, x):
        x = self.relu1(self.conv1(x))
        x = self.maxpool1(x)
        x = self.relu2(self.conv2(x))
        x = self.maxpool2(x)
        x = self.relu3(self.conv3(x))
        x = self.relu4(self.conv4(x))
        x = self.relu5(self.conv5(x))
        x = self.maxpool3(x)
        x = x.view(x.size(0), -1)
        # x = self.dropout(x)
        x = self.relu6(self.linear1(x))
        # x = self.dropout(x)
        x = self.relu7_last(self.linear2(x))
        x = self.linear3(x)
        # x = self.bn8(x)
        # x = self.relu8_last(x)
        return x        
    def extract_feature(self, x):
        x = self.conv1(x)
        feat1 = self.relu1(x)
        x = self.maxpool1(feat1)
        x = self.conv2(x)
        feat2 = self.relu2(x)
        x = self.maxpool2(feat2)
        x = self.conv3(x)
        feat3 = self.relu3(x)
        x = self.conv4(feat3)
        feat4 = self.relu4(x)
        x = self.conv5(feat4)
        feat5 = self.relu5(x)
        x = self.maxpool3(feat5)
        x = x.view(x.size(0), -1)
        x = self.linear1(x)
        feat6 = self.relu6(x)
        x = self.linear2(feat6)
        feat7 = self.relu7_last(x)
        if hasattr(self,'relu8_last'):
            if isinstance(self.relu8_last,nn.ReLU) or isinstance(self.relu8_last,Relu_bound) : 
                x =  self.linear3(feat7)
                feat8 = self.relu8_last(x)
                out = feat8
                return [feat8], out
        else:
            out = self.linear3(feat7)
            return [feat7], out

        
###########################################





import copy
import math
import os.path
from typing import Dict, Optional, Tuple, Type
from rrelu.utils.metric import accuracy
from torchpack import distributed as dist
from tqdm import tqdm
from rrelu.pytorchfi.weight_error_models import multi_weight_inj_fixed,multi_weight_inj_float,multi_weight_inj_int
from rrelu.relu_bound.bound_relu import Relu_bound
from rrelu.pytorchfi.core import FaultInjection
from rrelu.utils.distributed import DistributedMetric
import torch
import torch.nn as nn
import torchvision.datasets as datasets
import torchvision.transforms as transforms
from rrelu.models.vgg import make_layers,cfg
from rrelu.models import (
   Lenet,
   VGG,
   AlexNet_model,
   ResNet50,
   VGG16,
   LeNet_cifar,
   AlexNet_cifar100,
   VGG16_cifar100,
   ResNet50_cifar100,
)
from rrelu.relu_bound import (
    bounded_relu_zero,
    bounded_relu_tresh,
    bounded_relu_fitact,
    bounded_hyrelu_proact

)

from rrelu.search_bound import (
    FtClipAct_bounds,
    Ranger_bounds,
    fitact_bounds,
    proact_bounds

)
__all__ = ["build_data_loader", "build_model", "replace_act","find_bounds","eval_fault","eval"]

def eval_fault(model:nn.Module,data_loader_dict, fault_rate,iterations=500,bitflip=None,total_bits = 32 , n_frac = 16 , n_int = 15 )-> Dict:
    inputs, classes = next(iter(data_loader_dict['val'])) 
    pfi_model = FaultInjection(model, 
                            inputs.shape[0],
                            input_shape=[inputs.shape[1],inputs.shape[2],inputs.shape[3]],
                            layer_types=[torch.nn.Conv2d, torch.nn.Linear ,Relu_bound],
                            total_bits= total_bits,
                            n_frac = n_frac, 
                            n_int = n_int, 
                            use_cuda=True,
                            )
    print(pfi_model.print_pytorchfi_layer_summary())
    test_criterion = nn.CrossEntropyLoss().cuda()

    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()

    pfi_model.original_model.eval()
    with torch.no_grad():
        with tqdm(
            total= iterations,
            desc="Eval",
            disable=not dist.is_master(),
        ) as t:
            for i in range(iterations):
                if bitflip=='float':
                    corrupted_model = multi_weight_inj_float(pfi_model,fault_rate)
                elif bitflip=='fixed':    
                    corrupted_model = multi_weight_inj_fixed(pfi_model,fault_rate)
                elif bitflip =="int":
                    corrupted_model = multi_weight_inj_int (pfi_model,fault_rate)
                    # corrupted_model = multi_weight_inj_int(pfi_model,fault_rate)          
                for images, labels in data_loader_dict["val"]:
                    images, labels = images.cuda(), labels.cuda()
                    output = corrupted_model(images)
                    loss = test_criterion(output, labels)
                    val_loss.update(loss, images.shape[0])
                    acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                    val_top5.update(acc5[0], images.shape[0])
                    val_top1.update(acc1[0], images.shape[0])
                    
                ####        
                t.set_postfix(
                    {
                        "loss": val_loss.avg.item(),
                        "top1": val_top1.avg.item(),
                        "top5": val_top5.avg.item(),
                        "#samples": val_top1.count.item(),
                        "batch_size": images.shape[0],
                        "img_size": images.shape[2],
                        "fault_rate": fault_rate,
                    }
                )
                t.update()
                # pfi_model.original_model = corrupted_model    
        val_results = {
            "val_top1": val_top1.avg.item(),
            "val_top5": val_top5.avg.item(),
            "val_loss": val_loss.avg.item(),
            "fault_rate": fault_rate,
        }
    return val_results

def eval(model: nn.Module, data_loader_dict) -> Dict:

    test_criterion = nn.CrossEntropyLoss().cuda()

    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()

    model.eval()
    with torch.no_grad():
        with tqdm(
            total=len(data_loader_dict["val"]),
            desc="Eval",
            disable=not dist.is_master(),
        ) as t:
            for images, labels in data_loader_dict["val"]:
                images, labels = images.cuda(), labels.cuda()
                # compute output
                output = model(images)
                loss = test_criterion(output, labels)
                val_loss.update(loss, images.shape[0])
                acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                val_top5.update(acc5[0], images.shape[0])
                val_top1.update(acc1[0], images.shape[0])

                t.set_postfix(
                    {
                        "loss": val_loss.avg.item(),
                        "top1": val_top1.avg.item(),
                        "top5": val_top5.avg.item(),
                        "#samples": val_top1.count.item(),
                        "batch_size": images.shape[0],
                        "img_size": images.shape[2],
                    }
                )
                t.update()

    val_results = {
        "val_top1": val_top1.avg.item(),
        "val_top5": val_top5.avg.item(),
        "val_loss": val_loss.avg.item(),
    }
    return val_results


def build_data_loader(
    dataset: str,
    image_size: int,
    batch_size: int,
    n_worker: int = 8,
    data_path: Optional[str] = None,
    num_replica: Optional[int] = None,
    rank: Optional[int] = None,
) -> Tuple[Dict, int]:
    # build dataset
    dataset_info_dict = {
        "imagenet21k_winter_p": (
            os.path.expanduser("./dataset/imagenet21k_winter_p"),
            10450,
        ),
        "imagenet": (os.path.expanduser("./dataset/imagenet"), 1000),
        "car": (os.path.expanduser("./dataset/fgvc/stanford_car"), 196),
        "flowers102": (os.path.expanduser("./dataset/fgvc/flowers102"), 102),
        "food101": (os.path.expanduser("./dataset/fgvc/food101"), 101),
        "cub200": (os.path.expanduser("./dataset/fgvc/cub200"), 200),
        "pets": (os.path.expanduser("./dataset/fgvc/pets"), 37),
        "cifar10": (os.path.expanduser("./dataset/cifar10"), 10),
         "cifar100": (os.path.expanduser("./dataset/cifar100"), 100),
         "mnist":(os.path.expanduser("./dataset/mnist"), 10)
    }
    assert dataset in dataset_info_dict, f"Do not support {dataset}"

    data_path = data_path or dataset_info_dict[dataset][0]
    n_classes = dataset_info_dict[dataset][1]

    # build datasets
    if dataset=="mnist":
        train_dataset  =datasets.MNIST(
            os.path.join(data_path, "MNIST"),
            train=True,
            download=True,
            transform=transforms.Compose([
                transforms.ToTensor(),
                
                
            ]
            )
        )
        val_dataset = datasets.MNIST(
            os.path.join(data_path, "MNIST"),
            train=False,
            download=True,
            transform=transforms.Compose([
                 transforms.ToTensor(),
            ]     
            )
        )
    elif dataset=="cifar10":
        train_dataset  =datasets.CIFAR10(
            os.path.join(data_path, "CIFAR10"),
            train=True,
            download=True,
            transform=transforms.Compose([
                transforms.RandomCrop(32,padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                
                
            ]
            )
        )
        val_dataset = datasets.CIFAR10(
            os.path.join(data_path, "CIFAR10"),
            train=False,
            download=True,
            transform=transforms.Compose([
                 transforms.ToTensor(),
            ]     
            )
        )
    elif dataset=="cifar100":
        train_dataset  =datasets.CIFAR100(
            os.path.join(data_path, "CIFAR100"),
            train=True,
            download=True,
            transform=transforms.Compose([
                transforms.RandomCrop(32,padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                
            ]
            )
        )
        val_dataset = datasets.CIFAR100(
            os.path.join(data_path, "CIFAR100"),
            train=False,
            download=True,
            transform=transforms.Compose([
                 transforms.ToTensor(),
            ]     
            )
        )
    else:    
        train_dataset = datasets.ImageFolder(
            os.path.join(data_path, "train"),
            transforms.Compose(
                [
                    transforms.RandomResizedCrop(image_size),
                    transforms.RandomHorizontalFlip(),
                    transforms.ColorJitter(brightness=32.0 / 255.0, saturation=0.5),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                    ),
                ]
            ),
        )

        val_dataset = datasets.ImageFolder(
            os.path.join(data_path, "val"),
            transforms.Compose(
                [
                    transforms.Resize(int(math.ceil(image_size / 0.875))),
                    transforms.CenterCrop(image_size),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                    ),
                ]
            ),
        )

    sub_train_dataset = copy.deepcopy(train_dataset)  # used for resetting bn statistics
    sub_train_dataset.transform = val_dataset.transform
    if len(sub_train_dataset) > 16000:
        g = torch.Generator()
        g.manual_seed(937162211)
        rand_indexes = torch.randperm(len(sub_train_dataset), generator=g).tolist()
        rand_indexes = rand_indexes[:3000] # for alexnet and  vgg use 3000 for ftclip 1000
        if  dataset=="cifar10":
            sub_train_dataset.data = [
                sub_train_dataset.data[idx] for idx in rand_indexes
            ]
            sub_train_dataset.targets = [
                sub_train_dataset.targets[idx] for idx in rand_indexes
            ]
        elif  dataset=="cifar100":
            sub_train_dataset.data = [
                sub_train_dataset.data[idx] for idx in rand_indexes
            ]    
            sub_train_dataset.targets = [
                sub_train_dataset.targets[idx] for idx in rand_indexes
            ]
        elif  dataset=="mnist":
            sub_train_dataset.data = [
                sub_train_dataset.data[idx] for idx in rand_indexes
            ]    
            sub_train_dataset.targets = [
                sub_train_dataset.targets[idx] for idx in rand_indexes
            ]    
        else:   
            sub_train_dataset.samples = [
                sub_train_dataset.samples[idx] for idx in rand_indexes
            ]
            

        # build data loader
    if num_replica is None:
        train_loader = torch.utils.data.DataLoader(
            dataset=train_dataset,
            batch_size=batch_size,
            shuffle=True,
            num_workers=n_worker,
            pin_memory=True,
            drop_last=True,
        )
        sub_train_loader = torch.utils.data.DataLoader(
            dataset=sub_train_dataset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=n_worker,
            pin_memory=True,
            drop_last=True,
        )
        valid_loader = torch.utils.data.DataLoader(
            dataset=val_dataset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=n_worker,
            pin_memory=True,
            drop_last=False,
        )
    else:
        train_loader = torch.utils.data.DataLoader(
            dataset=train_dataset,
            batch_size=batch_size,
            sampler=torch.utils.data.distributed.DistributedSampler(
                train_dataset, num_replica, rank
            ),
            num_workers=n_worker,
            pin_memory=False,
            drop_last=True,
        )
        sub_train_loader = torch.utils.data.DataLoader(
            dataset=sub_train_dataset,
            batch_size=batch_size,
            sampler=torch.utils.data.distributed.DistributedSampler(
                sub_train_dataset, num_replica, rank
            ),
            num_workers=n_worker,
            pin_memory=False,
            drop_last=True,
            shuffle=False,
        )
        valid_loader = torch.utils.data.DataLoader(
            dataset=val_dataset,
            batch_size=batch_size,
            sampler=torch.utils.data.distributed.DistributedSampler(
                val_dataset, num_replica, rank
            ),
            num_workers=n_worker,
            pin_memory=False,
            drop_last=False,
            shuffle=False,
        )

    # prefetch sub_train
    sub_train_loader = [data for data in sub_train_loader]

    data_loader_dict = {
        "train": train_loader,
        "val": valid_loader,
        "sub_train": sub_train_loader,
    }

    return data_loader_dict, n_classes


def build_model(
    name: str,
    n_classes=10,
    dropout_rate=0.0,
    **kwargs,
) -> nn.Module:

    model_dict = {
        "lenet": Lenet,
        "lenet_cifar10": LeNet_cifar,
        "vgg16": VGG16,
        "resnet50": ResNet50,
        "alexnet": AlexNet_model,
        "alexnet_cifar100" : AlexNet_cifar100, 
        "vgg16_cifar100":VGG16_cifar100,
        "resnet50_cifar100": ResNet50_cifar100,
    }

    name = name.split("-")
    if len(name) > 1:
        kwargs["width_mult"] = float(name[1])
    name = name[0]

    return model_dict[name](n_classes=n_classes, dropout_rate=dropout_rate, **kwargs)


def find_bounds(model:nn.Module, data_loader, name:str,bound_type:str,bitflip:str):
    search_bounds_dict={
        "ranger" : Ranger_bounds,
        "ftclip" : FtClipAct_bounds,
        'fitact' : fitact_bounds,
        'proact'  : proact_bounds
    }
    return search_bounds_dict[name](model,data_loader,bound_type=bound_type,bitflip=bitflip)

def replace_act_all(model:nn.Module,relu_bound,bounds,tresh,alpha=None, name='')->nn.Module:
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU) and 'last' not in name1:
                name_ = name1 + name
                if alpha == None:
                    # print(name_)
                    model._modules[name1] = relu_bound(bounds[name_].detach(),tresh[name_].detach(),alpha)
                else:
                    model._modules[name1] = relu_bound(bounds[name_].detach(),tresh[name_].detach(),alpha[name_].detach())    
            elif isinstance(layer,nn.ReLU) and 'last' in name1:
                name_ = name1 + name
                if alpha == None:
                    # print(name_)
                    model._modules[name1] = relu_bound(bounds[name_].detach(),tresh[name_].detach(),alpha,k=-20.0)
                else:
                    model._modules[name1] = relu_bound(bounds[name_].detach(),tresh[name_].detach(),alpha[name_].detach(),k=-20.0) 
                    
        else:
            name+=name1
            replace_act_all(layer,relu_bound,bounds,tresh,alpha,name)               
    return model  
def replace_act(model:nn.Module, name_relu_bound:str, name_serach_bound:str,data_loader,bound_type:str,bitflip:str)->nn.Module:
    replace_act_dict={
        'zero' : bounded_relu_zero,
        'tresh': bounded_relu_tresh,
        'fitact': bounded_relu_fitact,
        'proact' : bounded_hyrelu_proact
    }
    bounds,tresh,alpha = find_bounds(copy.deepcopy(model),data_loader,name_serach_bound,bound_type,bitflip) 
    print(bounds)
    model = replace_act_all(model,replace_act_dict[name_relu_bound],bounds,tresh,alpha,name='')
    return model                
                         
    






import sys
import torch.nn as nn
import torch
import copy
import numpy as np
import warnings
import sys;
from rrelu.relu_bound.bound_proact import bounded_hyrelu_proact
from rrelu.relu_bound.bound_zero import bounded_relu_zero
import os
import argparse
from typing import Dict, Optional
from rrelu.relu_bound.bound_relu import Relu_bound
from rrelu.pytorchfi.weight_error_models import multi_weight_inj_float,multi_weight_inj_fixed,multi_weight_inj_int
from rrelu.utils.metric import accuracy,AverageMeter
from rrelu.utils.lr_scheduler import CosineLRwithWarmup
from rrelu.utils.distributed import DistributedMetric
import random
from rrelu.pytorchfi.core import FaultInjection
import torch.nn.functional as F
import time
from tqdm import tqdm
from torchpack import distributed as dist
parser = argparse.ArgumentParser()
parser.add_argument("--path", type=str, metavar="DIR", help="run directory",default="pretrained_models/lenet_mnist")
parser.add_argument("--base_batch_size", type=int, default=128)
parser.add_argument("--manual_seed", type=int, default=0)
parser.add_argument(
    "--gpu", type=str, default=None
)  # used in single machine experiments

def Ranger_bounds_proact(model:nn.Module, train_loader, device="cuda", bound_type='layer',bitflip = 'float'):
    model.eval()
    iteration = True 
    results={}
    tresh={}
    relu_hooks(model,name='')      
    for data, label in train_loader['sub_train']:
        data = data.to(device)
        label = label.to(device)
        model = model.to(device)
        output = model(data)
        if iteration:
            for key, val in activation.items():
                results[key] = val
                tresh[key] = val
            iteration = False

        for key, val in activation.items():
            prev_max = torch.max(results[key],dim=0)[0]
            prev_mean = torch.mean(tresh[key],dim=0)
            curr_max = torch.max(activation[key],dim=0)[0]
            curr_mean = torch.mean(activation[key],dim=0)
            results[key] = torch.maximum(prev_max,curr_max)
            tresh[key] = torch.minimum(prev_mean,curr_mean)   
    
    return results,tresh,None
def eval(model: nn.Module, data_loader_dict) :

    test_criterion = nn.CrossEntropyLoss().cuda()

    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()

    model.eval()
    with torch.no_grad():
        with tqdm(
            total=len(data_loader_dict["val"]),
            desc="Eval",
            disable=not dist.is_master(),
        ) as t:
            for images, labels in data_loader_dict["val"]:
                images, labels = images.cuda(), labels.cuda()
                # compute output
                output = model(images)
                loss = test_criterion(output, labels)
                val_loss.update(loss, images.shape[0])
                acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                val_top5.update(acc5[0], images.shape[0])
                val_top1.update(acc1[0], images.shape[0])

                t.set_postfix(
                    {
                        "loss": val_loss.avg.item(),
                        "top1": val_top1.avg.item(),
                        "top5": val_top5.avg.item(),
                        "#samples": val_top1.count.item(),
                        "batch_size": images.shape[0],
                        "img_size": images.shape[2],
                    }
                )
                t.update()

    val_results = {
        "val_top1": val_top1.avg.item(),
        "val_top5": val_top5.avg.item(),
        "val_loss": val_loss.avg.item(),
    }
    return val_results


activation={}
def get_activation(name):
    def hook(model, input, output):
        activation[name] = output
    return hook

def relu_hooks(model:nn.Module,name=''):
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU):
                name_ = name1 + name
                layer.register_forward_hook(get_activation(name_)) 
        else:
            name+=name1
            relu_hooks(layer,name)

def replace_act_all(model:nn.Module,bounds,tresh,name='')->nn.Module:
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU) and 'last' not in name1:
                name_ = name1 + name
                if tresh==None:
                    model._modules[name1] = bounded_hyrelu_proact(bounds[name_].detach(),tresh)   
                else:    
                    model._modules[name1] = bounded_hyrelu_proact(bounds[name_].detach(),tresh[name_].detach())  
            elif isinstance(layer,nn.ReLU) and 'last'  in name1:
                name_ = name1 + name
                if tresh==None:
                    model._modules[name1] = bounded_hyrelu_proact(bounds[name_].detach(),tresh,k=-20)   
                else:    
                    model._modules[name1] = bounded_hyrelu_proact(bounds[name_].detach(),tresh[name_].detach(),k=-20.0)  

        else:
            name+=name1
            replace_act_all(layer,bounds,tresh,name)               
    return model  
  

def eval_fault(model:nn.Module,data_loader_dict, fault_rate,iterations=2000,bitflip=None,total_bits = 32 , n_frac = 16 , n_int = 15 )-> Dict:
    inputs, classes = next(iter(data_loader_dict['val'])) 
    pfi_model = FaultInjection(model, 
                            inputs.shape[0],
                            input_shape=[inputs.shape[1],inputs.shape[2],inputs.shape[3]],
                            layer_types=[torch.nn.Conv2d, torch.nn.Linear ,Relu_bound],
                            total_bits= total_bits,
                            n_frac = n_frac, 
                            n_int = n_int, 
                            use_cuda=True,
                            )
    print(pfi_model.print_pytorchfi_layer_summary())
    test_criterion = nn.CrossEntropyLoss().cuda()

    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()

    pfi_model.original_model.eval()
    with torch.no_grad():
        with tqdm(
            total= iterations,
            desc="Eval",
            disable=not dist.is_master(),
        ) as t:
            for i in range(iterations):
                if bitflip=='float':
                    corrupted_model = multi_weight_inj_float(pfi_model,fault_rate)
                elif bitflip=='fixed':    
                    corrupted_model = multi_weight_inj_fixed(pfi_model,fault_rate)
                elif bitflip =="int":
                    corrupted_model = multi_weight_inj_int (pfi_model,fault_rate)
                    # corrupted_model = multi_weight_inj_int(pfi_model,fault_rate)    
                for images, labels in data_loader_dict["val"]:
                    images, labels = images.cuda(), labels.cuda()
                    output = corrupted_model(images)
                    # print(output)
                    loss = test_criterion(output, labels)
                    val_loss.update(loss, images.shape[0])
                    acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                    val_top5.update(acc5[0], images.shape[0])
                    val_top1.update(acc1[0], images.shape[0])
                    
                t.set_postfix(
                    {
                        "loss": val_loss.avg.item(),
                        "top1": val_top1.avg.item(),
                        "top5": val_top5.avg.item(),
                        "#samples": val_top1.count.item(),
                        "batch_size": images.shape[0],
                        "img_size": images.shape[2],
                        "fault_rate": fault_rate,
                    }
                )
                t.update()
                # pfi_model.original_model = corrupted_model    
        val_results = {
            "val_top1": val_top1.avg.item(),
            "val_top5": val_top5.avg.item(),
            "val_loss": val_loss.avg.item(),
            "fault_rate": fault_rate,
        }
    return val_results
from rrelu.models import (

   ResNet50,
)
def build_model(
    name: str,
    n_classes=10,
    dropout_rate=0.0,
    **kwargs,
) -> nn.Module:

    model_dict = {

        "resnet50": ResNet50,
    }

    name = name.split("-")
    if len(name) > 1:
        kwargs["width_mult"] = float(name[1])
    name = name[0]

    return model_dict[name](n_classes=n_classes, dropout_rate=dropout_rate, **kwargs)
def load_state_dict_from_file(file: str) -> Dict[str, torch.Tensor]:
    checkpoint = torch.load(file, map_location="cpu")
    if "state_dict" in checkpoint:
        checkpoint = checkpoint["state_dict"]
    return checkpoint
def proact_bounds(model:nn.Module, train_loader, device="cuda", bound_type='layer', bitflip='float'):
    model.eval()
    original_model  = copy.deepcopy(model)
    results,tresh,_ =  Ranger_bounds_proact(copy.deepcopy(model),train_loader,device,bound_type,bitflip) # FtClipAct_bounds(copy.deepcopy(model),teacher_model,train_loader,device,bound_type,bitflip)
    len_relu = len(results)
    if bound_type =="layer":
        for i,(key, val) in enumerate(results.items()):
                if i<len_relu - 1:
                    results[key] = torch.max(val)  
                    tresh[key] = torch.min(tresh[key]) 
    print(results)        
    model = replace_act_all(model,results,tresh)
    torch.save(model.state_dict(), "temp_{}_{}_{}.pth".format(bound_type,bitflip,original_model.__class__.__name__))      
    eval(model,train_loader)
    warnings.filterwarnings("ignore")
    args, opt = parser.parse_known_args()
    if args.gpu is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    if not torch.distributed.is_initialized():
        dist.init()
    torch.backends.cudnn.benchmark = True
    torch.cuda.set_device(dist.local_rank())
    os.makedirs(args.path, exist_ok=True)
    torch.manual_seed(args.manual_seed)
    torch.cuda.manual_seed_all(args.manual_seed)
    random.seed(args.manual_seed)
    np.random.seed(args.manual_seed)
    for name, param in model.named_parameters():
        if np.any([key in name for key in ["weight", "norm","bias"]]):
            param.requires_grad=False
        else:
            print(name)
            param.requires_grad=False      
    # model = nn.parallel.DistributedDataParallel(
    #     model.cuda(), device_ids=[dist.local_rank()]
    # )
    # for name, param in model.named_parameters():
    #     if np.any([key in name for key in ["weight", "norm","bias"]]):
    #         param.requires_grad=False
    #     else:
    #         print(name)
    #         param.requires_grad=False 
     
    weight_decay_list =[4e-1,4e-2,4e-3,4e-4,4e-5,4e-6,4e-7,4e-8,4e-9,4e-10,4e-11]
    model = train(model,original_model,train_loader,weight_decay_list,bound_type,bitflip)
    for name, param in model.named_parameters():
        if np.any([key in name for key in ["weight", "norm","bias"]]):
            param.requires_grad=False
        else:
            print(name)
            param.requires_grad=True   
    bounds_dict = {}
    keys=[]
    i=0
   
    for key,val in results.items():
        keys.append(key)
    # print(keys)    
    for name, param in model.named_parameters():
        if param.requires_grad:
            if np.any([key in name for key in ["bounds_param"]]):
                # print(param)
                bounds_dict[keys[i]]=param
                i+=1
    print(bounds_dict)            
    return bounds_dict,tresh,None



def distillation_loss(feat_s, feat_t,inputs,device):
    # loss = loss * ((teacher_hidden_representation > student_hidden_representation) | (teacher_hidden_representation > 0)).float()
    cosine_loss =  nn.CosineSimilarity()
    # KL_loss = nn.KLDivLoss()
    return torch.mean(cosine_loss(feat_s.view(feat_s.size(0),-1), feat_t.view(feat_t.size(0),-1))) #oss.sum()
    # return KL_loss(feat_t,feat_s)
def kl_loss(a,b):
    loss = -a*b + torch.log(b+1e-5)*b
    return loss
def cross_entropy_loss_with_soft_target(pred, soft_target):
    logsoftmax = nn.LogSoftmax()
    return torch.mean(torch.sum(-soft_target * logsoftmax(pred), 1))
def L1_reg(model):
    nweights = 0
    for name,param in model.named_parameters():
        if param.requires_grad==True:
            if 'bounds'  in name:
                nweights = nweights + param.numel()
    L1_term = torch.tensor(0., requires_grad=True)
    for name, param in model.named_parameters():
        if param.requires_grad==True:
            if 'bounds'  in name:
                weights_sum = torch.sum(torch.abs(param))
                L1_term = L1_term + weights_sum
    L1_term = L1_term / nweights
    return L1_term


def train(model,original_model,data_provider,weight_decay_list,base_lr=0.01,warmup_epochs=5,n_epochs=5 , treshold=torch.tensor(1.5),bound_type = "layer" , bitflip = "fixed"):
    val_info_dict = eval(model, data_provider)
    best_acc =torch.tensor(val_info_dict["val_top1"])
    for name, param in reversed(list(model.named_parameters())):
        if np.any([key in name for key in ["weight", "norm","bias"]]):
            param.requires_grad=False
            continue
        else:
            print(param.nelement() ,name)
            param.requires_grad=True 
            if param.nelement() > 3 : 
                base_lr = 0.001
                n_epochs = 100
                warmup_epochs=5
            else:
                base_lr = 0.01   
                n_epochs = 10
                warmup_epochs=5
        for wd in weight_decay_list:
            params_with_wd=[]
            for name, param in model.named_parameters():
                if param.requires_grad:
                    params_with_wd.append(param)
            net_params = [
                {
                    "params": params_with_wd,
                    "weight_decay": wd,
                },
            ]
            optimizer = torch.optim.AdamW(
                net_params,
                lr=base_lr * dist.size(),
                weight_decay=wd
            )
            # build lr scheduler
            lr_scheduler = CosineLRwithWarmup(
                optimizer,
                warmup_epochs * len(data_provider['train']),
                base_lr,
                n_epochs * len(data_provider['train']),
            )
            # train criterion
            train_criterion = nn.CrossEntropyLoss()
            # test criterion 
            test_criterion = nn.CrossEntropyLoss()
            # init
            
            start_epoch = 0
            for epoch in range(
                start_epoch,
                n_epochs
                + warmup_epochs,
            ):
                train_info_dict = train_one_epoch(
                    model,
                    original_model,
                    data_provider,
                    epoch,
                    optimizer,
                    train_criterion,
                    lr_scheduler,
                )

            val_info_dict = eval(model, data_provider)
            if torch.abs(best_acc - val_info_dict["val_top1"]) <=treshold:
                print(wd)
                torch.save(model.state_dict(), "temp_{}_{}_{}.pth".format(bound_type,bitflip,original_model.__class__.__name__))
                param.requires_grad = False
                break
            else:
                model.load_state_dict(torch.load("temp_{}_{}_{}.pth".format(bound_type,bitflip,original_model.__class__.__name__)))  
    return model



def train_one_epoch(
    model: nn.Module,
    original_model,
    data_provider,
    epoch: int,
    optimizer,
    criterion,
    lr_scheduler,
):
    train_loss = DistributedMetric()
    train_top1 = DistributedMetric()
    model.train()
    data_provider['train'].sampler.set_epoch(epoch)

    data_time = AverageMeter()
    with tqdm(
        total=len(data_provider["train"]) ,
        desc="Train Epoch #{}".format(epoch + 1),
        disable=not dist.is_master(),
    ) as t:
        end = time.time()         
        for _, (images, labels) in enumerate(data_provider['train']):
            data_time.update(time.time() - end)
            images, labels = images.cuda(), labels.cuda()
            l2_bounds = 0.0
            for name, param in model.named_parameters():
                if param.requires_grad==True:
                    if "bounds" in name:
                        l2_bounds += torch.mean(torch.pow(param, 2))
            optimizer.zero_grad()
            with torch.no_grad():
                teacher_output = original_model(images).detach()
                # feat_t,_ = original_model.extract_feature(images)
                teacher_logits = F.softmax(teacher_output, dim=1)
            # feat_s,_ = model.extract_feature(images)    
            nat_logits = model(images)
            kd_loss = cross_entropy_loss_with_soft_target(
                        nat_logits,teacher_logits
                    )
            # k_loss = distillation_loss(feat_s,feat_t,images,'cuda')
            nat_logits = model(images)
            loss =  0.8 * criterion(nat_logits,labels) +0.1 * kd_loss  +  4e-8 * l2_bounds #+ 0.1 * k_loss #
            loss.backward()
            # for name,par in model.named_parameters():
            #     if par.requires_grad == True:
            #         print(par.grad)
            top1 = accuracy(nat_logits, labels, topk=(1,))[0][0]
            optimizer.step()
            lr_scheduler.step()

            train_loss.update(loss, images.shape[0])
            train_top1.update(top1, images.shape[0])

            t.set_postfix(
                {
                    "loss": train_loss.avg.item(),
                    "top1": train_top1.avg.item(),
                    "batch_size": images.shape[0],
                    "img_size": images.shape[2],
                    "lr": optimizer.param_groups[0]["lr"],
                    "data_time": data_time.avg,
                }
            )
            t.update()

            end = time.time()
    return {
        "train_top1": train_top1.avg.item(),
        "train_loss": train_loss.avg.item(),
    }



from typing import List, Union

import numpy as np
import torch

__all__ = ["accuracy", "AverageMeter"]


def accuracy(
    output: torch.Tensor, target: torch.Tensor, topk=(1,)
) -> List[torch.Tensor]:
    """Computes the precision@k for the specified values of k."""
    maxk = max(topk)
    batch_size = target.shape[0]

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.reshape(1, -1).expand_as(pred))
    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


class AverageMeter(object):
    """Computes and stores the average and current value.

    Copied from: https://github.com/pytorch/examples/blob/master/imagenet/main.py
    """

    def __init__(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val: Union[torch.Tensor, np.ndarray, float, int], n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


import torch.nn as nn
import torch.nn.functional as F
class Lenet(nn.Module):
    def __init__(self,n_classes = 10 , dropout_rate=0.0):
        super(Lenet, self).__init__()
        self.conv1 = nn.Conv2d(1, 6, 5)
        self.relu1 = nn.ReLU()
        self.pool1 = nn.MaxPool2d(2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.relu2 = nn.ReLU()
        self.pool2 = nn.MaxPool2d(2)
        self.fc1 = nn.Linear(256, 120)
        self.relu3 = nn.ReLU()
        self.fc2 = nn.Linear(120, 84)
        self.relu4 = nn.ReLU()
        self.fc3 = nn.Linear(84, n_classes)

    def forward(self, x):
        y = self.conv1(x)
        y = self.relu1(y)
        y = self.pool1(y)
        y = self.conv2(y)
        y = self.relu2(y)
        y = self.pool2(y)
        y = y.view(y.shape[0], -1)
        y = self.fc1(y)
        y = self.relu3(y)
        y = self.fc2(y)
        y = self.relu4(y)
        y = self.fc3(y)
        return y
    def extract_feature(self, x, preReLU=False):
        feat1 = self.conv1(x)
        x = self.relu1(feat1)
        x = self.pool1(x)
        feat2 = self.conv2(x)
        x = self.relu2(feat2)
        x = self.pool2(x)
        x = x.view(x.size(0), -1)
        feat3 = self.fc1(x)
        x = self.relu3(feat3)
        feat4 = self.fc2(x)
        x = self.relu4(feat4)
        out = self.fc3(x)
        if not preReLU:
            feat1 = F.relu(feat1)
            feat2 = F.relu(feat2)
            feat3 = F.relu(feat3)
            feat4 = F.relu(feat4)
        return [feat1, feat2, feat3,feat4], out
class LeNet_cifar(nn.Module):
    def __init__(self,n_classes = 10 , dropout_rate=0.0):
        super(LeNet_cifar, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.relu1 = nn.ReLU()
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.relu2 = nn.ReLU()
        self.fc1   = nn.Linear(16*5*5, 120)
        self.relu3 = nn.ReLU()
        self.fc2   = nn.Linear(120, 84)
        self.relu4 = nn.ReLU()
        self.fc3   = nn.Linear(84, n_classes)

    def forward(self, x):
        out = self.relu1(self.conv1(x))
        out = F.max_pool2d(out, 2)
        out = self.relu2(self.conv2(out))
        out = F.max_pool2d(out, 2)
        out = out.view(out.size(0), -1)
        out = self.relu3(self.fc1(out))
        out = self.relu4(self.fc2(out))
        out = self.fc3(out)
        return out
    def extract_feature(self, x, preReLU=False):
        feat1 = self.conv1(x)
        x = self.relu1(feat1)
        x = F.max_pool2d(x, 2)
        feat2 = self.conv2(x)
        x = self.relu2(feat2)
        x = F.max_pool2d(x, 2)
        x = x.view(x.size(0), -1)
        feat3 = self.fc1(x)
        x = self.relu3(feat3)
        feat4 = self.fc2(x)
        x = self.relu4(feat4)
        out = self.fc3(x)
        if not preReLU:
            feat1 = F.relu(feat1)
            feat2 = F.relu(feat2)
            feat3 = F.relu(feat3)
            feat4 = F.relu(feat4)
        return [feat1, feat2, feat3,feat4], out


import torch
import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F
import numpy
class AlexNet_cifar100(nn.Module):

  def __init__(self, n_classes=100,dropout_rate=0.0):
    super(AlexNet_cifar100, self).__init__()
    self.features = nn.Sequential(
      nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1),
      nn.ReLU(inplace=True),
      nn.MaxPool2d(kernel_size=3, stride=2),
      nn.Conv2d(64, 192, kernel_size=3, stride=1, padding=1),
      nn.ReLU(inplace=True),
      nn.MaxPool2d(kernel_size=3, stride=2),
      nn.Conv2d(192, 384, kernel_size=3, stride=1, padding=1),
      nn.ReLU(inplace=True),
      nn.Conv2d(384, 256, kernel_size=3, stride=1, padding=1),
      nn.ReLU(inplace=True),
      nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
      nn.ReLU(inplace=True),
      nn.MaxPool2d(kernel_size=3, stride=2),
    )
    self.classifier = nn.Sequential(
      nn.Dropout(),
      nn.Linear(256 * 1 * 1, 4096),
      nn.ReLU(inplace=True),
      nn.Dropout(),
      nn.Linear(4096, 4096),
      nn.ReLU(inplace=True),
      nn.Linear(4096, n_classes),
    )

  def forward(self, x):
    x = self.features(x)
    x = torch.flatten(x, 1)
    x = self.classifier(x)
    return x


"""pytorchfi.error_models provides different error models out-of-the-box for use."""

import random
from fxpmath import Fxp
import rrelu.pytorchfi.core as core 
from rrelu.pytorchfi.util import random_value
import logging
import torch
import numpy as np
def random_weight_location(pfi, layer: int = -1):
    if layer == -1:
        layer = random.randint(0, pfi.get_total_layers() - 1)

    dim = pfi.get_weights_dim(layer)
    shape = pfi.get_weights_size(layer)

    dim0_shape = shape[0]
    k = random.randint(0, dim0_shape - 1)
    if dim > 1:
        dim1_shape = shape[1]
        dim1_rand = random.randint(0, dim1_shape - 1)
    if dim > 2:
        dim2_shape = shape[2]
        dim2_rand = random.randint(0, dim2_shape - 1)
    else:
        dim2_rand = None
    if dim > 3:
        dim3_shape = shape[3]
        dim3_rand = random.randint(0, dim3_shape - 1)
    else:
        dim3_rand = None

    return ([layer], [k], [dim1_rand], [dim2_rand], [dim3_rand])


# Weight Perturbation Models
def random_weight_inj(
    pfi, corrupt_layer: int = -1, min_val: int = -1, max_val: int = 1
):
    layer, k, c_in, kH, kW = random_weight_location(pfi, corrupt_layer)
    faulty_val = [random_value(min_val=min_val, max_val=max_val)]

    return pfi.declare_weight_fault_injection(
        layer_num=layer, k=k, dim1=c_in, dim2=kH, dim3=kW, value=faulty_val
    )


def zero_func_rand_weight(pfi: core.FaultInjection):
    layer, k, c_in, kH, kW = random_weight_location(pfi)
    return pfi.declare_weight_fault_injection(
        function=_zero_rand_weight, layer_num=layer, k=k, dim1=c_in, dim2=kH, dim3=kW
    )
def bit2float(b, device=torch.device('cpu'), num_e_bits=8, num_m_bits=23, bias=127.):
    expected_last_dim = num_m_bits + num_e_bits + 1
    dtype = torch.float32

    s = torch.index_select(b, -1, torch.arange(0, 1, device=device))
    e = torch.index_select(b, -1, torch.arange(1, 1 + num_e_bits, device=device))
    m = torch.index_select(b, -1, torch.arange(1 + num_e_bits,
                                             1 + num_e_bits + num_m_bits, device=device))
    
    ## SIGN BIT    
    out = (torch.pow((torch.ones(1, device=device) * -1), s)).squeeze(-1).type(dtype)
    
    ## EXPONENT BIT
    exponents = -torch.arange(-(num_e_bits - 1.), 1., device=device)
    exponents = exponents.repeat(b.shape[:-1] + (1,))
    e_decimal = torch.sum(e * torch.pow(torch.ones(1, device=device) * 2, exponents), dim=-1) - bias
    out *= torch.pow(torch.ones(1, device=device) * 2, e_decimal)
    
    ## MANTISSA
    matissa = (torch.pow((torch.ones(1, device=device) * 2), -torch.arange(1., num_m_bits + 1., device=device))).repeat(m.shape[:-1] + (1,))
    
    out *= 1. + torch.sum(m * matissa, dim=-1)
    
    #Correcting 0 and inf
    out[torch.abs(out) < 5e-30] = 0
    out[out == float("inf")] = 10e30
    out[out == float("-inf")] = -10e30
    out = out.detach()
    
    return out



def float2bit(f, device=torch.device('cpu'), num_e_bits=8, num_m_bits=23, bias=127., dtype=torch.float32):
    ## SIGN BIT
    s = torch.sign(f)
    f = f * s
    
    # turn sign into sign-bit
    s = (s * (-1) + 1.) * 0.5
    s = s.unsqueeze(-1)
    s[s == 0.5] = 0

    ## EXPONENT BIT
    e_scientific = torch.floor(torch.log2(f))
    e_decimal = e_scientific + bias
    e = integer2bit(e_decimal, num_bits=num_e_bits, device=device)
    e = torch.nan_to_num(e)
    e[e == float("inf")] = 0
    e[e == float("-inf")] = 0
    # print(e)
    ## MANTISSA
    int_precision = 1024
    m1 = integer2bit(f - f % 1, num_bits=int_precision, device=device)
    m2 = remainder2bit(f % 1, num_bits=bias, device=device)
    
    m = torch.cat([m1, m2], dim=-1)
    dtype = f.type()
    
    idx = torch.arange(num_m_bits, device=device).unsqueeze(0).type(dtype) + (float(int_precision) - e_scientific).unsqueeze(-1)
    # print(idx)

    idx[idx == float("inf")] = 0
    idx[idx == float("-inf")] = 0
    idx = torch.nan_to_num(idx)
    
    idx = idx.long()
    
    m = torch.gather(m, dim=-1, index=idx)
    
    out = torch.cat([s, e, m], dim=-1).type(dtype)
    out = out.detach()
    # print(out.shape)
    return out


def remainder2bit(remainder, num_bits=127, device=torch.device('cpu')):
    dtype = remainder.type()
    exponent_bits = torch.arange(num_bits, device=device).type(dtype)
    exponent_bits = exponent_bits.repeat(remainder.shape + (1,))
    out = (remainder.unsqueeze(-1) * 2 ** exponent_bits) % 1
    return torch.floor(2 * out)


def integer2bit(integer, num_bits=32, device=torch.device('cpu')):
    dtype = integer.type()
    exponent_bits = -torch.arange(-(num_bits - 1), 1, device=device).type(dtype)
    exponent_bits = exponent_bits.repeat(integer.shape + (1,))
    out = integer.unsqueeze(-1) / 2 ** exponent_bits
    return (out - (out % 1)) % 2


def quantized_int2bit(qint_tensor, qbit, device=torch.device('cpu')):       #converts a tensor of integers into binary
    qint_tensor = qint_tensor.type(torch.int)
    mask = 2 ** torch.arange(qbit - 1, -1, -1).to(qint_tensor.device, qint_tensor.dtype)
    return qint_tensor.unsqueeze(-1).bitwise_and(mask).ne(0).to(torch.float)


def quantized_bit2int(qbinary_tensor, qbit, device=torch.device('cpu')):    #converts binary into integers tensor
    mask = 2 ** torch.arange(qbit - 1, -1, -1).to(qbinary_tensor.device, qbinary_tensor.dtype)
    mask[0] = -mask[0]      #MSB is sign bit
    return torch.sum(mask * qbinary_tensor, -1)

def _zero_rand_weight(data, location):
    new_data = data[location] * 0
    return new_data

def twos_comp(val, bits):
    if (val & (1 << (bits - 1))) != 0:
            val = val - (1 << bits)
    return val

def twos_comp_shifted(val, nbits):
    return (1 << nbits) + val if val < 0 else twos_comp(val, nbits)


def flip_bit_signed(orig_value, max_value, bit_pos):
        # quantum value
        save_type = orig_value.dtype
        total_bits = 31
        logging.info(f"Original Value: {orig_value}")

        quantum = int((orig_value / max_value) * ((2.0 ** (total_bits - 1))))
        twos_comple = twos_comp_shifted(quantum, total_bits)  # signed
        logging.info(f"Quantum: {quantum}")
        logging.info(f"Twos Couple: {twos_comple}")

        # binary representation
        bits = bin(twos_comple)[2:]
        logging.info(f"Bits: {bits}")

        # sign extend 0's
        temp = "0" * (total_bits - len(bits))
        bits = temp + bits
        if len(bits) != total_bits:
            raise AssertionError
        logging.info(f"Sign extend bits {bits}")

        # flip a bit
        # use MSB -> LSB indexing
        if bit_pos >= total_bits:
            raise AssertionError

        bits_new = list(bits)
        bit_loc = total_bits - bit_pos - 1
        if bits_new[bit_loc] == "0":
            bits_new[bit_loc] = "1"
        else:
            bits_new[bit_loc] = "0"
        bits_str_new = "".join(bits_new)
        logging.info(f"New bits: {bits_str_new}")

        # GPU contention causes a weird bug...
        if not bits_str_new.isdigit():
            logging.info("Error: Not all the bits are digits (0/1)")

        # convert to quantum
        if not bits_str_new.isdigit():
            raise AssertionError
        new_quantum = int(bits_str_new, 2)
        out = twos_comp(new_quantum, total_bits)
        logging.info(f"Out: {out}")

        # get FP equivalent from quantum
        new_value = out * ((2.0 ** (-1 * (total_bits - 1))) * max_value)
        logging.info(f"New Value: {new_value}")

        return torch.tensor(new_value, dtype=save_type)

    
 


def bit_flip_weight_fixed(data, location , bit = None , total_bits = None ,n_frac = None , n_int = None):
    orig_value = data[location].item() 
    
    total_bits = total_bits   
    x = Fxp(orig_value, True, n_word=total_bits,n_frac=n_frac,n_int=n_int)
    bits = x.bin()
     
    bits_new = list(bits)
    
    bit_pos = bit   
    
    bit_loc = total_bits - bit_pos - 1
    if bits_new[bit_loc] == '0':
        bits_new[bit_loc] = '1'
    else:
        bits_new[bit_loc] = '0'
    bits_str_new = "".join(bits_new)
    if not bits_str_new.isdigit():
        print("Error: Not all the bits are digits (0/1)")

    # convert to quantum
    assert bits_str_new.isdigit()
    FXP_value = x("0b"+bits_str_new)
    new_value = FXP_value.astype(float)

    return torch.tensor(new_value)
def int2bin(input, num_bits):
    '''
    convert the signed integer value into unsigned integer (2's complement equivalently).
    Note that, the conversion is different depends on number of bit used.
    '''
    output = input.clone()
    if num_bits == 1: # when it is binary, the conversion is different
        output = output/2 + .5
    elif num_bits > 1:
        output[input.lt(0)] = 2**num_bits + output[input.lt(0)]

    return output


def bin2int(input, num_bits):
    '''
    convert the unsigned integer (2's complement equivantly) back to the signed integer format
    with the bitwise operations. Note that, in order to perform the bitwise operation, the input
    tensor has to be in the integer format.
    '''
    if num_bits == 1:
        output = input*2-1
    elif num_bits > 1:
        mask = 2**(num_bits - 1) - 1
        output = -(input & ~mask) + (input & mask)
    return output


def bit_flip_weight_int(data, location , bit = None , total_bits = None ,n_frac = None , n_int = None):
    orig_value = torch.tensor([data[location].item()],dtype=torch.int) 
    # print(orig_value)
    twos_comple = twos_comp_shifted(orig_value, total_bits)  # signed
    bits = bin(twos_comple)[2:]
    temp = "0" * (total_bits - len(bits))
    bits = temp + bits
    if len(bits) != total_bits:
        raise AssertionError
        
    # flip a bit
    # use MSB -> LSB indexing
    if bit >= total_bits:
        raise AssertionError

    bits_new = list(bits)
    bit_loc = total_bits - bit - 1
    if bits_new[bit_loc] == "0":
        bits_new[bit_loc] = "1"
    else:
        bits_new[bit_loc] = "0"
    bits_str_new = "".join(bits_new)
    

    # GPU contention causes a weird bug...
    if not bits_str_new.isdigit():
        logging.info("Error: Not all the bits are digits (0/1)")

    # convert to quantum
    if not bits_str_new.isdigit():
        raise AssertionError
    new_quantum = int(bits_str_new, 2)
    out = twos_comp(new_quantum, total_bits)
    return torch.tensor(out)

def bit_flip_weight_IEEE(data, location , bit = None , total_bits = None ,n_frac = None , n_int = None):
    orig_value = torch.tensor([data[location].item()]) 
    total_bits = total_bits    
    bit_pos = bit
    # print(bit_pos)
    # binary representation
    bits= float2bit(orig_value)
    # print(bits)
    bit_loc = total_bits - bit_pos - 1
    # if bits[0,bit_loc] == 0:
    #     bits[0,bit_loc] = 1
    # else:
    #     bits[0,bit_loc] = 0
    bits[0,bit_loc] = torch.logical_xor(bits[0,bit_loc],torch.tensor(1.0)).float()    
    new_value = bit2float(bits)
    # print(orig_value,new_value)

    return torch.tensor(new_value[0])


def multi_weight_inj_int(pfi, sdc_p=1e-5, function1=bit_flip_weight_int,function2=bit_flip_weight_IEEE):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    for layer_idx in range(pfi.get_total_layers()):
        shape = list(pfi.get_weights_size(layer_idx))
        shape_bias = list(pfi.get_bias_size(layer_idx))
        dim_bias = len(shape_bias)
        dim_len = len(shape)  
        shape.extend([1 for i in range(4 - len(shape))])
        nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
        shape_bias.extend([1 for i in range(4 - len(shape_bias))])
        if shape_bias[0] !=None : 
           nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
        if nunmber_fault_weight !=0: 
            k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
            dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
            dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
            dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
            dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
        if shape_bias[0]!=None:
            if nunmber_fault_bias!=0:
                k_b = torch.randint(shape[0],(nunmber_fault_bias,), device='cuda')
                dim1_b = torch.randint(shape[1],(nunmber_fault_bias,), device='cuda')
                dim2_b = torch.randint(shape[2],(nunmber_fault_bias,), device='cuda')
                dim3_b = torch.randint(shape[3],(nunmber_fault_bias,), device='cuda')
                dim4_b = torch.randint(total_bits,(nunmber_fault_bias,), device='cuda')
        for fault in range(nunmber_fault_weight):
            idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
            for i in range(dim_len + 1):
                corrupt_idx[i].append(idx[i])
            for i in range(dim_len + 1, 5): 
                corrupt_idx[i].append(None)  
            corrupt_bit_idx.append(dim4_w[fault]) 
        if shape_bias[0]!= None:                                                                       
            for fault in range(nunmber_fault_bias):
                idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
                for i in range(dim_bias + 1):
                    corrupt_idx_bias[i].append(idx[i])
                for i in range(dim_bias + 1, 5): 
                    corrupt_idx_bias[i].append(None)  
                corrupt_bit_idx_bias.append(dim4_b[fault])       
                                                                     
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = None,
        n_int = None
    )



def multi_weight_inj_float(pfi, sdc_p=1e-5, function1=bit_flip_weight_IEEE,function2=bit_flip_weight_IEEE):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    for layer_idx in range(pfi.get_total_layers()):
        shape = list(pfi.get_weights_size(layer_idx))
        shape_bias = list(pfi.get_bias_size(layer_idx))
        dim_bias = len(shape_bias)
        dim_len = len(shape)  
        shape.extend([1 for i in range(4 - len(shape))])
        nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
        shape_bias.extend([1 for i in range(4 - len(shape_bias))])
        if shape_bias[0] !=None : 
           nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
        if nunmber_fault_weight !=0:   
            k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
            dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
            dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
            dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
            dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
        if shape_bias[0]!=None:
            if nunmber_fault_bias!=0:
                k_b = torch.randint(shape[0],(nunmber_fault_bias,), device='cuda')
                dim1_b = torch.randint(shape[1],(nunmber_fault_bias,), device='cuda')
                dim2_b = torch.randint(shape[2],(nunmber_fault_bias,), device='cuda')
                dim3_b = torch.randint(shape[3],(nunmber_fault_bias,), device='cuda')
                dim4_b = torch.randint(total_bits,(nunmber_fault_bias,), device='cuda')
        for fault in range(nunmber_fault_weight):
            idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
            for i in range(dim_len + 1):
                corrupt_idx[i].append(idx[i])
            for i in range(dim_len + 1, 5): 
                corrupt_idx[i].append(None)  
            corrupt_bit_idx.append(dim4_w[fault]) 
        if shape_bias[0]!= None:                                                                       
            for fault in range(nunmber_fault_bias):
                idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
                for i in range(dim_bias + 1):
                    corrupt_idx_bias[i].append(idx[i])
                for i in range(dim_bias + 1, 5): 
                    corrupt_idx_bias[i].append(None)  
                corrupt_bit_idx_bias.append(dim4_b[fault])                                                           
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = None,
        n_int = None
    )



def multi_weight_inj_fixed(pfi, sdc_p=1e-5, function1=bit_flip_weight_fixed,function2=bit_flip_weight_fixed):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    for layer_idx in range(pfi.get_total_layers()):
        shape = list(pfi.get_weights_size(layer_idx))
        shape_bias = list(pfi.get_bias_size(layer_idx))
        dim_bias = len(shape_bias)
        dim_len = len(shape)  
        shape.extend([1 for i in range(4 - len(shape))])
        nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
        shape_bias.extend([1 for i in range(4 - len(shape_bias))])
        if shape_bias[0] !=None : 
           nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
        if nunmber_fault_weight !=0:      
            k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
            dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
            dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
            dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
            dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
        if shape_bias[0]!=None:
            if nunmber_fault_bias!=0:
                k_b = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
                dim1_b = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
                dim2_b = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
                dim3_b = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
                dim4_b = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
        for fault in range(nunmber_fault_weight):
            idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
            for i in range(dim_len + 1):
                corrupt_idx[i].append(idx[i])
            for i in range(dim_len + 1, 5): 
                corrupt_idx[i].append(None)  
            corrupt_bit_idx.append(dim4_w[fault]) 
        if shape_bias[0]!= None:                                                                       
            for fault in range(nunmber_fault_bias):
                idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
                for i in range(dim_bias + 1):
                    corrupt_idx_bias[i].append(idx[i])
                for i in range(dim_bias + 1, 5): 
                    corrupt_idx_bias[i].append(None)  
                corrupt_bit_idx_bias.append(dim4_b[fault])           
                                                                     
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = n_frac, 
        n_int = n_int, 
    )




def binaryOfFraction(fraction):
    binary = str()
    while (fraction):
        fraction *= 2
        if (fraction >= 1):
            int_part = 1
            fraction -= 1
        else:
            int_part = 0
        binary += str(int_part)
    return binary
def floatingPoint(real_no):
    sign_bit = 0
    if (real_no < 0):
        sign_bit = 1
    real_no = abs(real_no)
    int_str = bin(int(real_no))[2:]
    if real_no - int(real_no) !=0.0:
        fraction_str = binaryOfFraction(real_no - int(real_no))
    else:
        fraction_str=''    
    if int(real_no)==0:
        ind = int_str.index('0')
    else:
        ind = int_str.index('1')
    exp_str = bin((len(int_str) - ind - 1) + 127)[2:]
    # print(exp_str)
    # print(fraction_str)
    mant_str = int_str[ind + 1:] + fraction_str
    mant_str = mant_str + ('0' * (23 - len(mant_str)))
    # print(mant_str)
    ieee_32 = str(sign_bit) + exp_str  + mant_str
    return ieee_32

if __name__=="__main__":
    a = torch.tensor([127])
    print(bit_flip_weight_int(a,0,8,8))



import sys
import torch
import numpy as np
import torch
import copy
import torch.nn as nn 
from rrelu.pytorchfi.core import FaultInjection
from rrelu.utils.distributed import DistributedMetric
from tqdm import tqdm
from torchpack import distributed as dist
from rrelu.utils import load_state_dict_from_file
from rrelu.utils.metric import accuracy
from rrelu.pytorchfi.weight_error_models import bit_flip_weight_IEEE,bit_flip_weight_fixed,bit_flip_weight_int
from rrelu.relu_bound.bound_zero import bounded_relu_zero 
from rrelu.relu_bound.bound_relu import Relu_bound
import random
activation={}
def get_activation(name):
    def hook(model, input, output):
        activation[name] = output
    return hook

def relu_hooks(model:nn.Module,name=''):
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU):
                name_ = name1 + name
                layer.register_forward_hook(get_activation(name_)) 
        else:
            name+=name1
            relu_hooks(layer,name)

def replace_act_all(model:nn.Module,bounds,tresh,name='')->nn.Module:
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU):
                name_ = name1 + name
                model._modules[name1] = bounded_relu_zero(bounds[name_],tresh[name_])
        else:
            name+=name1
            replace_act_all(layer,bounds,tresh,name)               
    return model  
def replace_act(model:nn.Module,bounds,key,tresh,name='')->nn.Module:
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU)or isinstance(layer,Relu_bound):
                name_ = name1 + name
                if name_==key:
                    model._modules[name1] = bounded_relu_zero(bounds,tresh)
        else:
            name+=name1
            replace_act(layer,bounds,key,tresh,name)               
    return model  
def FtClipAct_bounds(model:nn.Module, data_loader, device="cuda",bound_type='layer',bitflip='float',N=4, M = 3,fault_rates=[1e-7,1e-6,3e-6,1e-5,3e-5], delta_b=0.2):
    layer_idx = 0
    model.eval()
    model.cuda()
    
    iteration = True 
    max_results={}
    bounds_results={}
    tresh_results={}
    relu_hooks(model)       
    for data, label in data_loader['sub_train']:
        data = data.to(device)
        label = label.to(device)
        model = model.to(device)
        output = model(data)
        if iteration:
            for key, val in activation.items():
                max_results[key] = val
                bounds_results[key]= val
                tresh_results[key] = val
            iteration = False
        for key, val in activation.items():
            prev_max = torch.max(max_results[key],dim=0)[0]
            prev_mean = torch.mean(tresh_results[key],dim=0)
            curr_mean = torch.mean(activation[key],dim=0)
            curr_max = torch.max(activation[key],dim=0)[0]
            tresh_results[key] = torch.minimum(prev_mean,curr_mean)
            max_results[key] = torch.maximum(prev_max,curr_max)                
    if bound_type=="neuron":
        raise AssertionError("FTClip Act doesnt work with neuron wise")
    else:            
        for key, val in max_results.items():
            max_results[key] = torch.max(val).detach()
            tresh_results[key] = torch.mean(tresh_results[key]).detach()    
    for key,val in max_results.items():
        print(key,val)
        counter = 1
        i = 1 
        act_max = val.item()
        tresh_val = tresh_results[key].item()
        while counter<=N:
            if i==1:
                S = torch.tensor([0,act_max])
                AUC,T = AUC_Calculation(S,copy.deepcopy(model),fault_rates,data_loader,key,device,tresh_val,layer_idx,bitflip)
                S,T_result =Interval_Search(T,AUC)
            else:
                AUC, T = AUC_Calculation(S, copy.deepcopy(model) , fault_rates, data_loader,key ,device,tresh_val,layer_idx,bitflip)
                S,T_result =Interval_Search(T,AUC)
            Delta=[]
            for j in range(3):
                Delta.append(torch.abs(AUC[j+1]-AUC[j]))
            Delta = torch.tensor(Delta)
            if torch.max(Delta)<=delta_b and counter>=M:
                bounds_results[key] = T_result
                # model = replace_act(model,T_result,key,tresh_val)
                break
            counter+=1
            i+=1
        bounds_results[key] = T_result
        print("the fial bounded is {}\n".format(T_result))
        # model = replace_act(model,T_result,key,tresh_val)
        # print(model)
        layer_idx+=1
    for key,val in bounds_results.items():
        torch.save(bounds_results[key],"{}_{}_{}_{}.pt".format(model.__class__.__name__,bound_type,bitflip,key))
    return bounds_results,tresh_results,None  




def AUC_Calculation(S,model,fault_rates,data_loader,key,device,tresh,layer_idx,bitflip):
    model.eval()
    T1 = torch.min(S)
    T2 = T1 + (torch.max(S)-torch.min(S))/3
    T3 = T2 + (torch.max(S)-torch.min(S))/3
    T4 = torch.max(S)
    T=[T1,T2,T3,T4]
    AUC=[]
    orig_model = copy.deepcopy(model)
    for i in range(4):
        model = replace_act(model,T[i],key,tresh) 
        auc = auc_compute(model , fault_rates, data_loader,device,layer_idx,bitflip)
        AUC.append(auc)
        model = copy.deepcopy(orig_model)
    return torch.tensor(AUC),torch.tensor(T)




      

def auc_compute(model,fault_rates,data_loader,device,layer_idx,bitflip):
    model.eval()
    acc_list=[]
    acc = 0 
    orig_model = copy.deepcopy(model)
    for fault_rate in fault_rates:
        acc = accuracy_vs_faultrate(model,fault_rate,data_loader,layer_idx,bitflip)
        # print(acc)
        acc_list.append(acc)
        acc=0
        model = copy.deepcopy(orig_model)
    fault_rates = np.array([0.2,0.4,0.6,0.8,1])
    acc_list = np.array(acc_list)
    auc = np.trapz(acc_list,fault_rates)
    return  auc
def Interval_Search(T,AUC):
    '''
    interval serach in ftclip act paper to serahc better treshold
    ------------------------------------------------------------
    T : list of the bounds
    AUC : AUC
    '''
    index = torch.argmax(AUC)
    if index==3:
        S = torch.tensor([T[2],T[3]])
    elif index==0:
        S = torch.tensor([T[0], T[1]])
    else:
        S = torch.tensor([T[index-1],T[index+1]])
    T_prime = T[index]
    return S,T_prime





def accuracy_vs_faultrate(model,fault_rate,data_loader,layer_idx,bitflip,fault_simulation_epochs=20):
    inputs, classes = next(iter(data_loader['sub_train']))
    pfi_model = FaultInjection(copy.deepcopy(model), 
                inputs.shape[0],
                input_shape=[inputs.shape[1],inputs.shape[2],inputs.shape[3]],
                layer_types=[torch.nn.Conv2d, torch.nn.Linear],
                use_cuda=True,
                )
    model.cuda()
    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()
    test_criterion = nn.CrossEntropyLoss().cuda()
    pfi_model.original_model.eval()
    with torch.no_grad():
        for i in range(fault_simulation_epochs):  
            if bitflip=='float':
                    corrupted_model = multi_weight_inj_float(pfi_model,fault_rate,layer_idx=layer_idx)
            elif bitflip=='fixed':    
                    corrupted_model = multi_weight_inj_fixed(pfi_model,fault_rate,layer_idx=layer_idx)
            elif bitflip =="int":
                    corrupted_model = multi_weight_inj_int (pfi_model,fault_rate,layer_idx=layer_idx)
            for images, labels in data_loader["sub_train"]:
                images, labels = images.cuda(), labels.cuda()
                # compute output

                output = corrupted_model(images)
                loss = test_criterion(output, labels)
                val_loss.update(loss, images.shape[0])
                acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                val_top5.update(acc5[0], images.shape[0])
                val_top1.update(acc1[0], images.shape[0])
        # print(val_top1.avg.item())
    return val_top1.avg.item()




########################## Fault Injection ##################################
def multi_weight_inj_int(pfi, sdc_p=1e-5, function1=bit_flip_weight_int,function2=bit_flip_weight_IEEE,layer_idx=0):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    # for layer_idx in range(layer_idx_max):
    shape = list(pfi.get_weights_size(layer_idx))
    shape_bias = list(pfi.get_bias_size(layer_idx))
    dim_bias = len(shape_bias)
    dim_len = len(shape)  
    shape.extend([1 for i in range(4 - len(shape))])
    nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
    shape_bias.extend([1 for i in range(4 - len(shape_bias))])
    if shape_bias[0] !=None : 
        nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
    if nunmber_fault_weight !=0: 
        k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
        dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
        dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
        dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
        dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
    if shape_bias[0]!=None:
        if nunmber_fault_bias!=0:
            k_b = torch.randint(shape[0],(nunmber_fault_bias,), device='cuda')
            dim1_b = torch.randint(shape[1],(nunmber_fault_bias,), device='cuda')
            dim2_b = torch.randint(shape[2],(nunmber_fault_bias,), device='cuda')
            dim3_b = torch.randint(shape[3],(nunmber_fault_bias,), device='cuda')
            dim4_b = torch.randint(total_bits,(nunmber_fault_bias,), device='cuda')
    for fault in range(nunmber_fault_weight):
        idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
        for i in range(dim_len + 1):
            corrupt_idx[i].append(idx[i])
        for i in range(dim_len + 1, 5): 
            corrupt_idx[i].append(None)  
        corrupt_bit_idx.append(dim4_w[fault]) 
    if shape_bias[0]!= None:                                                                       
        for fault in range(nunmber_fault_bias):
            idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
            for i in range(dim_bias + 1):
                corrupt_idx_bias[i].append(idx[i])
            for i in range(dim_bias + 1, 5): 
                corrupt_idx_bias[i].append(None)  
            corrupt_bit_idx_bias.append(dim4_b[fault])       
                                                                     
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = None,
        n_int = None
    )



def multi_weight_inj_float(pfi, sdc_p=1e-5, function1=bit_flip_weight_IEEE,function2=bit_flip_weight_IEEE,layer_idx=0):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    # for layer_idx in range(layer_idx_max):
    shape = list(pfi.get_weights_size(layer_idx))
    shape_bias = list(pfi.get_bias_size(layer_idx))
    dim_bias = len(shape_bias)
    dim_len = len(shape)  
    shape.extend([1 for i in range(4 - len(shape))])
    nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
    shape_bias.extend([1 for i in range(4 - len(shape_bias))])
    if shape_bias[0] !=None : 
        nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
    if nunmber_fault_weight !=0:   
        k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
        dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
        dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
        dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
        dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
    if shape_bias[0]!=None:
        if nunmber_fault_bias!=0:
            k_b = torch.randint(shape[0],(nunmber_fault_bias,), device='cuda')
            dim1_b = torch.randint(shape[1],(nunmber_fault_bias,), device='cuda')
            dim2_b = torch.randint(shape[2],(nunmber_fault_bias,), device='cuda')
            dim3_b = torch.randint(shape[3],(nunmber_fault_bias,), device='cuda')
            dim4_b = torch.randint(total_bits,(nunmber_fault_bias,), device='cuda')
    for fault in range(nunmber_fault_weight):
        idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
        for i in range(dim_len + 1):
            corrupt_idx[i].append(idx[i])
        for i in range(dim_len + 1, 5): 
            corrupt_idx[i].append(None)  
        corrupt_bit_idx.append(dim4_w[fault]) 
    if shape_bias[0]!= None:                                                                       
        for fault in range(nunmber_fault_bias):
            idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
            for i in range(dim_bias + 1):
                corrupt_idx_bias[i].append(idx[i])
            for i in range(dim_bias + 1, 5): 
                corrupt_idx_bias[i].append(None)  
            corrupt_bit_idx_bias.append(dim4_b[fault])                                                           
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = None,
        n_int = None
    )



def multi_weight_inj_fixed(pfi, sdc_p=1e-5, function1=bit_flip_weight_fixed,function2=bit_flip_weight_fixed,layer_idx=0):
    corrupt_idx = [[], [], [], [], []]
    corrupt_bit_idx = []
    corrupt_idx_bias = [[], [], [], [], []]
    corrupt_bit_idx_bias = []
    total_bits,n_frac,n_int = pfi.get_total_bits()
    # for layer_idx in range(layer_idx_max):
    shape = list(pfi.get_weights_size(layer_idx))
    shape_bias = list(pfi.get_bias_size(layer_idx))
    dim_bias = len(shape_bias)
    dim_len = len(shape)  
    shape.extend([1 for i in range(4 - len(shape))])
    nunmber_fault_weight = int(shape[0] * shape[1] * shape[2] * shape [3] * total_bits * sdc_p)
    shape_bias.extend([1 for i in range(4 - len(shape_bias))])
    if shape_bias[0] !=None : 
        nunmber_fault_bias = int(shape_bias[0] * shape_bias[1] * shape_bias[2] * shape_bias [3] * total_bits * sdc_p) 
    if nunmber_fault_weight !=0:      
        k_w = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
        dim1_w = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
        dim2_w = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
        dim3_w = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
        dim4_w = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
    if shape_bias[0]!=None:
        if nunmber_fault_bias!=0:
            k_b = torch.randint(shape[0],(nunmber_fault_weight,), device='cuda')
            dim1_b = torch.randint(shape[1],(nunmber_fault_weight,), device='cuda')
            dim2_b = torch.randint(shape[2],(nunmber_fault_weight,), device='cuda')
            dim3_b = torch.randint(shape[3],(nunmber_fault_weight,), device='cuda')
            dim4_b = torch.randint(total_bits,(nunmber_fault_weight,), device='cuda')
    for fault in range(nunmber_fault_weight):
        idx = [layer_idx, k_w[fault].item(), dim1_w[fault].item(), dim2_w[fault].item(), dim3_w[fault].item()]
        for i in range(dim_len + 1):
            corrupt_idx[i].append(idx[i])
        for i in range(dim_len + 1, 5): 
            corrupt_idx[i].append(None)  
        corrupt_bit_idx.append(dim4_w[fault]) 
    if shape_bias[0]!= None:                                                                       
        for fault in range(nunmber_fault_bias):
            idx = [layer_idx, k_b[fault].item(), dim1_b[fault].item(), dim2_b[fault].item(), dim3_b[fault].item()]
            for i in range(dim_bias + 1):
                corrupt_idx_bias[i].append(idx[i])
            for i in range(dim_bias + 1, 5): 
                corrupt_idx_bias[i].append(None)  
            corrupt_bit_idx_bias.append(dim4_b[fault])           
                                                                     
    return pfi.declare_weight_fault_injection(
        layer_num=[corrupt_idx[0],corrupt_idx_bias[0]],
        k=[corrupt_idx[1],corrupt_idx_bias[1]],
        dim1=[corrupt_idx[2],corrupt_idx_bias[2]],
        dim2=[corrupt_idx[3],corrupt_idx_bias[3]],
        dim3=[corrupt_idx[4],corrupt_idx_bias[4]],
        dim4 = [corrupt_bit_idx,corrupt_bit_idx_bias],
        function=[function1,function2],
        total_bits = total_bits,
        n_frac = n_frac, 
        n_int = n_int, 
    )




###############################################################################################



import sys
import torch.nn as nn
import torch
import sys;
activation={}
def get_activation(name):
    def hook(model, input, output):
        activation[name] = output
    return hook


def relu_hooks(model:nn.Module,name=''):
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU):
                name_ = name1 + name
                layer.register_forward_hook(get_activation(name_)) 
        else:
            name+=name1
            relu_hooks(layer,name)
             
def Ranger_bounds(model:nn.Module, train_loader, device="cuda", bound_type='layer',bitflip = 'float'):
    model.eval()
    iteration = True 
    results={}
    tresh={}
    relu_hooks(model,name='')      
    for data, label in train_loader['sub_train']:
        data = data.to(device)
        label = label.to(device)
        model = model.to(device)
        output = model(data)
        if iteration:
            for key, val in activation.items():
                results[key] = val
                tresh[key] = val
            iteration = False

        for key, val in activation.items():
            prev_max = torch.max(results[key],dim=0)[0]
            prev_mean = torch.mean(tresh[key],dim=0)
            curr_max = torch.max(activation[key],dim=0)[0]
            curr_mean = torch.mean(activation[key],dim=0)
            results[key] = torch.maximum(prev_max,curr_max)
            tresh[key] = torch.minimum(prev_mean,curr_mean)   
    
    if bound_type =="layer":
        for key, val in results.items():
            results[key] = torch.max(val)  
            tresh[key] = torch.min(tresh[key]) 
            
       
    return results,tresh,None


   





import torch
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound
class bounded_relu_tresh(nn.Module,Relu_bound):
    '''
    Bound the relu activatoin and back the values to treshold
    ------------------------------------
    bound : the bound for the activation
    -------------------------------------
    pytorch module with forward function
    '''
    def __init__(self, bounds ,tresh=None,alpha=None,k=-20):
        super().__init__()
        self.bounds = bounds
        self.tresh = tresh
    def forward(self, input):
        # input = torch.nan_to_num(input)
        output = torch.ones_like(input) * input
        output[torch.gt(input,self.bounds)] = self.tresh
        return torch.maximum(torch.tensor(0.0),output)       



import torch
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound
class bounded_relu_fitact(nn.Module,Relu_bound):
    def __init__(self,bounds,tresh=None,alpha = None,k=-20):
        super().__init__()
        bounds_param={}
        param_name= "bounds_param"
        self.tresh = None
        bounds_param[param_name] = nn.Parameter(data=bounds.cuda(), requires_grad=True)  
        self.k = k 
        for name, param in bounds_param.items():
            self.register_parameter(name, param) 
        self.bounds =  self.__getattr__("bounds_param")   
    def forward(self,input):
        # input = torch.nan_to_num(input)
        output = input - input * torch.sigmoid(-self.k * (input-self.__getattr__("bounds_param")))
        # print(output)
        return torch.maximum(torch.tensor(0.0),output)   




import torch
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound
class bounded_relu_zero(nn.Module,Relu_bound):
    '''
    Bound the relu activatoin and back the values to zero
    ------------------------------------
    bound : the bound for the activation
    --------------------------------------
    pytorch module with forward function
    '''
    def __init__(self, bounds,tresh = None,alpha=None , k = -20):
        super().__init__()
        self.bounds = bounds
        self.tresh = None
    def forward(self, input):
        # input = torch.nan_to_num(input)
        output = torch.ones_like(input) * input
        output[torch.gt(input,self.bounds)] = torch.tensor(0.0)
        return torch.maximum(torch.tensor(0.0),output)      



from rrelu.models.alexnet import *
from rrelu.models.lenet import *
from rrelu.models.vgg import *
from rrelu.models.resnet import *
from rrelu.models.alexnet_cifar100 import *
from rrelu.models.vgg_cifar100 import *
from rrelu.models.resnet_cifar100 import *



import math
from typing import Dict, List, Union

import torch
import torch.nn as nn
from torch.nn.modules.batchnorm import _BatchNorm

__all__ = ["init_modules", "load_state_dict"]


def init_modules(
    module: Union[nn.Module, List[nn.Module]], init_type="he_fout"
) -> None:
    init_params = init_type.split("@")
    if len(init_params) > 1:
        init_params = float(init_params[1])
    else:
        init_params = None

    if isinstance(module, list):
        for sub_module in module:
            init_modules(sub_module)
    else:
        for m in module.modules():
            if isinstance(m, nn.Conv2d):
                if init_type == "he_fout":
                    n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                    m.weight.data.normal_(0, math.sqrt(2.0 / n))
                elif init_type.startswith("kaiming_uniform"):
                    nn.init.kaiming_uniform_(m.weight, a=math.sqrt(init_params or 5))
                else:
                    nn.init.kaiming_uniform_(m.weight, a=math.sqrt(init_params or 5))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, _BatchNorm):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    m.bias.data.zero_()
            else:
                weight = getattr(m, "weight", None)
                bias = getattr(m, "bias", None)
                if isinstance(weight, torch.nn.Parameter):
                    nn.init.kaiming_uniform_(m.weight, a=math.sqrt(init_params or 5))
                if isinstance(bias, torch.nn.Parameter):
                    bias.data.zero_()


def load_state_dict(
    model: nn.Module, state_dict: Dict[str, torch.Tensor], strict=True
) -> None:
    current_state_dict = model.state_dict()
    for key in state_dict:
        if current_state_dict[key].shape != state_dict[key].shape:
            if strict:
                raise ValueError(
                    "%s shape mismatch (src=%s, target=%s)"
                    % (
                        key,
                        list(state_dict[key].shape),
                        list(current_state_dict[key].shape),
                    )
                )
            else:
                print(
                    "Skip loading %s due to shape mismatch (src=%s, target=%s)"
                    % (
                        key,
                        list(state_dict[key].shape),
                        list(current_state_dict[key].shape),
                    )
                )
        else:
            current_state_dict[key].copy_(state_dict[key])
    model.load_state_dict(current_state_dict)



import torch.nn as nn
import math
import torch.nn.functional as F
class VGG(nn.Module):
    '''
    VGG model 
    '''
    def __init__(self, n_classes=10,dropout_rate=0.0,features=None):
        super(VGG, self).__init__()
        self.features = features
        self.linear1 = nn.Linear(512, n_classes)
        # self.relu1 = nn.ReLU()
        # self.linear2 = nn.Linear(512, 512)
        # self.relu2_last= nn.ReLU()
        # self.linear3 = nn.Linear(512, n_classes)
        # self.relu = nn.ReLU()
         # Initialize weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                m.bias.data.zero_()


    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        logits = self.linear1(x)
        # x = self.relu1(x)
        # x = self.linear2(x)
        # x = self.relu2_last(x)
        # logits = self.linear3(x)
       
        return logits
    def extract_feature(self, x, preReLU=False):

        feat1 = self.features(x)
        x = feat1.view(x.size(0), -1)
        out = self.linear1(x)
        # feat2 = self.relu1(x)
        # x = self.linear2(feat2)
        # feat3 = self.relu2_last(x)
        return feat1, out

        if not preReLU:
            feat1 = F.relu(feat1)

        return [feat1], out

def make_layers(cfg, batch_norm=False):
    layers = []
    in_channels = 3
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU()]
            else:
                layers += [conv2d, nn.ReLU()]
            in_channels = v
    return nn.Sequential(*layers)


cfg = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 
          512, 512, 512, 512, 'M'],
}

class VGG16(VGG):
    def __init__(self, n_classes=10,dropout_rate=0.0,features=make_layers(cfg['D'],batch_norm=False)):
        super(VGG16,self).__init__(n_classes,dropout_rate,features)
##################################################################



import math
from typing import List

import torch
from torch.optim import Optimizer

__all__ = ["CosineLRwithWarmup"]


class CosineLRwithWarmup(torch.optim.lr_scheduler._LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        warmup_steps: int,
        warmup_lr: float,
        decay_steps: int,
        last_epoch: int = -1,
    ) -> None:
        self.warmup_steps = warmup_steps
        self.warmup_lr = warmup_lr
        self.decay_steps = decay_steps
        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        if self.last_epoch < self.warmup_steps:
            return [
                (base_lr - self.warmup_lr) * self.last_epoch / self.warmup_steps
                + self.warmup_lr
                for base_lr in self.base_lrs
            ]
        else:
            current_steps = self.last_epoch - self.warmup_steps
            return [
                0.5
                * base_lr
                * (1 + math.cos(math.pi * current_steps / self.decay_steps))
                for base_lr in self.base_lrs
            ]


import sys
import torch.nn as nn
import torch
import copy
import numpy as np
import warnings
import sys; 
from rrelu.relu_bound.bound_fitact import bounded_relu_fitact
import os
import argparse
from rrelu.utils.metric import accuracy,AverageMeter
from rrelu.utils.lr_scheduler import CosineLRwithWarmup
from rrelu.utils.distributed import DistributedMetric
from rrelu.search_bound.ranger import Ranger_bounds
import random
import time
from tqdm import tqdm
from torchpack import distributed as dist
parser = argparse.ArgumentParser()
parser.add_argument("--path", type=str, metavar="DIR", help="run directory",default="./pretrained_models/teachers/vgg16_bound_layer_relu_float/checkpoint")
parser.add_argument("--base_batch_size", type=int, default=128)
parser.add_argument("--manual_seed", type=int, default=0)
parser.add_argument(
    "--gpu", type=str, default=None
)  # used in single machine experiments


def eval(model: nn.Module, data_loader_dict) :

    test_criterion = nn.CrossEntropyLoss().cuda()

    val_loss = DistributedMetric()
    val_top1 = DistributedMetric()
    val_top5 = DistributedMetric()

    model.eval()
    with torch.no_grad():
        with tqdm(
            total=len(data_loader_dict["val"]),
            desc="Eval",
            disable=not dist.is_master(),
        ) as t:
            for images, labels in data_loader_dict["val"]:
                images, labels = images.cuda(), labels.cuda()
                # compute output
                output = model(images)
                loss = test_criterion(output, labels)
                val_loss.update(loss, images.shape[0])
                acc1, acc5 = accuracy(output, labels, topk=(1, 5))
                val_top5.update(acc5[0], images.shape[0])
                val_top1.update(acc1[0], images.shape[0])

                t.set_postfix(
                    {
                        "loss": val_loss.avg.item(),
                        "top1": val_top1.avg.item(),
                        "top5": val_top5.avg.item(),
                        "#samples": val_top1.count.item(),
                        "batch_size": images.shape[0],
                        "img_size": images.shape[2],
                    }
                )
                t.update()

    val_results = {
        "val_top1": val_top1.avg.item(),
        "val_top5": val_top5.avg.item(),
        "val_loss": val_loss.avg.item(),
    }
    return val_results


activation={}
def get_activation(name):
    def hook(model, input, output):
        activation[name] = output
    return hook

def relu_hooks(model:nn.Module,name=''):
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU):
                name_ = name1 + name
                layer.register_forward_hook(get_activation(name_)) 
        else:
            name+=name1
            relu_hooks(layer,name)

def replace_act(model:nn.Module,bounds,tresh,name='')->nn.Module:
    for name1,layer in model.named_children():
        if list(layer.children()) == []:
            if isinstance(layer,nn.ReLU) and 'last' not in name1:
                name_ = name1 + name
                model._modules[name1] = bounded_relu_fitact(bounds[name_].detach(),tresh[name_],-20)
            elif  isinstance(layer,nn.ReLU) and 'last' in name1:
                print("last relu layer")
                name_ = name1 + name
                model._modules[name1] = bounded_relu_fitact(bounds[name_].detach(),tresh[name_],-20)
        else:
            name+=name1
            replace_act(layer,bounds,tresh,name)               
    return model  
  




def fitact_bounds(model:nn.Module,train_loader, device="cuda", bound_type='layer',bitflip='float'):
    model.eval()
    results,tresh,_ = Ranger_bounds(copy.deepcopy(model),train_loader,device,bound_type,bitflip)
    # print(results['relu1'])
    model = replace_act(model,results,tresh)
    # print(model)
    warnings.filterwarnings("ignore")
    args, opt = parser.parse_known_args()
    if args.gpu is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    if not torch.distributed.is_initialized():
        dist.init()
    torch.backends.cudnn.benchmark = True
    torch.cuda.set_device(dist.local_rank())
    os.makedirs(args.path, exist_ok=True)
    torch.manual_seed(args.manual_seed)
    torch.cuda.manual_seed_all(args.manual_seed)
    random.seed(args.manual_seed)
    np.random.seed(args.manual_seed)
    for name, param in model.named_parameters():
        if np.any([key in name for key in ["weight", "norm","bias"]]):
            param.requires_grad=False
        else:
            param.requires_grad=True
    model = nn.parallel.DistributedDataParallel(
        model.cuda(), device_ids=[dist.local_rank()]
    )
    for name, param in model.named_parameters():
        if np.any([key in name for key in ["weight", "norm","bias"]]):
            param.requires_grad=False
        else:
            param.requires_grad=True

    model = train(model, train_loader, args.path)
    # print(eval(model,train_loader))
    bounds_dict = {}
    keys=[]
    i=0
   
    for key,val in results.items():
        keys.append(key)
    # print(keys)    
    for name, param in model.module.named_parameters():
        if param.requires_grad:
            if np.any([key in name for key in ["bounds_param"]]):
                # print(param)
                bounds_dict[keys[i]]=param
                i+=1
    return bounds_dict,bounds_dict,None

def train(
    model: nn.Module,
    data_provider,
    path: str,
    base_lr=0.001,
    warmup_epochs = 0 ,
    n_epochs = 150,
    weight_decay = 4e-9

):
    
    params_without_wd = []
    params_with_wd = []
    for name, param in model.named_parameters():
        if param.requires_grad:

            if np.any([key in name for key in ["bias", "norm"]]):
                params_without_wd.append(param)
            else:
                # print(name)
                params_with_wd.append(param)
    net_params = [
        {"params": params_without_wd, "weight_decay": 0},
        {
            "params": params_with_wd,
            "weight_decay": weight_decay,
        },
    ]
    # build optimizer
    optimizer = torch.optim.Adam(
        net_params,
        lr=base_lr * dist.size(),
    )
    # build lr scheduler
    lr_scheduler = CosineLRwithWarmup(
        optimizer,
        warmup_epochs * len(data_provider['train']),
        base_lr,
        n_epochs * len(data_provider['train']),
    )
    # train criterion
    train_criterion = nn.CrossEntropyLoss()
    # init
    best_val = 0.0
    start_epoch = 0
    checkpoint_path = os.path.join(path, "checkpoint")
    log_path = os.path.join(path, "logs")
    os.makedirs(checkpoint_path, exist_ok=True)
    os.makedirs(log_path, exist_ok=True)
    logs_writer = open(os.path.join(log_path, "exp.log"), "a")


    # start training
    for epoch in range(
        start_epoch,
        n_epochs
        + warmup_epochs,
    ):
        train_info_dict = train_one_epoch(
            model,
            data_provider,
            epoch,
            optimizer,
            train_criterion,
            lr_scheduler,
        )
    
        val_info_dict = eval(model, data_provider)
        is_best = val_info_dict["val_top1"] > best_val
        best_val = max(best_val, val_info_dict["val_top1"])
        # log
        epoch_log = f"[{epoch + 1 - warmup_epochs}/{n_epochs}]"
        epoch_log += f"\tval_top1={val_info_dict['val_top1']:.2f} ({best_val:.2f})"
        epoch_log += f"\ttrain_top1={train_info_dict['train_top1']:.2f}\tlr={optimizer.param_groups[0]['lr']:.2E}"
        if dist.is_master():
            logs_writer.write(epoch_log + "\n")
            logs_writer.flush()
        # save checkpoint
        checkpoint = {
            "state_dict": model.module.state_dict(),
            "epoch": epoch,
            "best_val": best_val,
            "optimizer": optimizer.state_dict(),
            "lr_scheduler": lr_scheduler.state_dict(),
        }
        if dist.is_master():
            torch.save(
                checkpoint,
                os.path.join(checkpoint_path, "checkpoint.pt"),
                _use_new_zipfile_serialization=False,
            )
            if is_best:
                torch.save(
                    checkpoint,
                    os.path.join(checkpoint_path, "best.pt"),
                    _use_new_zipfile_serialization=False,
                )
    
         
    return model




def train_one_epoch(
    model: nn.Module,
    data_provider,
    epoch: int,
    optimizer,
    criterion,
    lr_scheduler,

):
    train_loss = DistributedMetric()
    train_top1 = DistributedMetric()

    model.train()
    data_provider['train'].sampler.set_epoch(epoch)

    data_time = AverageMeter()
    with tqdm(
        total=len(data_provider["train"]),
        desc="Train Epoch #{}".format(epoch + 1),
        disable=not dist.is_master(),
    ) as t:
        end = time.time()
        for _, (images, labels) in enumerate(data_provider['train']):
            data_time.update(time.time() - end)
            images, labels = images.cuda(), labels.cuda()
            l2_bounds = 0.0
            for name, param in model.named_parameters():
                if param.requires_grad==True:
                    if "bounds" in name:
                        l2_bounds += torch.mean(torch.pow(param, 2))
            optimizer.zero_grad()
            output = model(images)
            loss = criterion(output, labels) + 4e-9 * l2_bounds
            loss.backward()
            # for par in model.parameters():
            #     print(par.grad)
            top1 = accuracy(output, labels, topk=(1,))[0][0]
            optimizer.step()
            lr_scheduler.step()

            train_loss.update(loss, images.shape[0])
            train_top1.update(top1, images.shape[0])

            t.set_postfix(
                {
                    "loss": train_loss.avg.item(),
                    "top1": train_top1.avg.item(),
                    "batch_size": images.shape[0],
                    "img_size": images.shape[2],
                    "lr": optimizer.param_groups[0]["lr"],
                    "data_time": data_time.avg,
                }
            )
            t.update()

            end = time.time()
    return {
        "train_top1": train_top1.avg.item(),
        "train_loss": train_loss.avg.item(),
    }






import torch
import torch.nn as nn
from rrelu.relu_bound.bound_relu import Relu_bound
import torch.nn.functional as F

class bounded_hyrelu_proact(nn.Module,Relu_bound):
    def __init__(self,bounds=None,tresh=None,alpha_param = None,k=-20):
        super().__init__()
        bounds_param={}
        param_name1= "bounds_param"
        self.tresh = tresh
        if tresh ==None:
            bounds_param[param_name1] = nn.Parameter(data=torch.zeros_like(bounds).cuda(), requires_grad=True) 
        else:
            bounds_param[param_name1] = nn.Parameter(data=bounds.cuda(), requires_grad=True) 
          
        self.k = k 
        for name, param in bounds_param.items():
            self.register_parameter(name, param)   
        self.bounds =  self.__getattr__("bounds_param")  
    def forward(self,input):
        # input = torch.nan_to_num(input)
        output =   input - input * torch.sigmoid(-self.k* (input-self.__getattr__("bounds_param"))) # + self.__getattr__("bounds_param")
        # print(output)
        return torch.maximum(torch.tensor(0.0),output)   
